def shell_app(app_connector, **kwargs):
    """
    Support Kwargs:
        user_name (str): Set user name to be displayed.
        assistant_name (dict): Set assistant name to be displayed.
    """
    if not app_connector.message_handler:
        raise Exception("No Message Handler, use `.set_message_handler()` to set one.")
    user_name = kwargs.get("user_name", "User")
    assistant_name = kwargs.get("assistant_name", "Assistant")
    message = ""
    chat_history = []
    while True:
        message = input(f"[{ user_name }]:")
        if message == "#exit":
            break
        # Define generator event handlers
        app_connector.on("delta", lambda data: { "yield": data })
        app_connector.on("done", lambda data: { "end": True })
        # Run message handler
        app_connector.run_message_handler(message, chat_history)
        print(f"[{ assistant_name }]:")
        buffer = ""
        try:
            for item in app_connector.data_generator.start():
                if item:
                    print(item, end="")
                    buffer += item
            chat_history.append({ "role": "user", "content": message })
            chat_history.append({ "role": "assistant", "content": buffer })
            app_connector.reset_data_generator()
            continue
        except StopIteration:
            app_connector.reset_data_generator()
            continue


import asyncio
import threading
import queue
import json
import json5

from ..utils import RuntimeCtx, RuntimeCtxNamespace, PluginManager, AliasManager, load_json, DataGenerator
from .._global import global_plugin_manager, global_settings

class Request(object):
    def __init__(
            self,
            *,
            parent_plugin_manager: object = global_plugin_manager,
            parent_request_runtime_ctx: object = None,
            parent_settings: object = global_settings,
        ):
        self.request_runtime_ctx = RuntimeCtx(parent = parent_request_runtime_ctx)
        self.settings = RuntimeCtx(parent = parent_settings)
        self.response_cache = {
            "prompt": {},
            "type": None,
            "reply": None,
        }
        self.response_generator = DataGenerator()
        # Plugin Manager
        self.plugin_manager = PluginManager(parent = parent_plugin_manager)
        # Namespace
        self.model = RuntimeCtxNamespace("model", self.settings, return_to = self)
        self.prompt_general = RuntimeCtxNamespace("prompt.general", self.request_runtime_ctx, return_to = self)
        self.prompt_role = RuntimeCtxNamespace("prompt.role", self.request_runtime_ctx, return_to = self)
        self.prompt_user_info = RuntimeCtxNamespace("prompt.user_info", self.request_runtime_ctx, return_to = self)
        self.prompt_abstract = RuntimeCtxNamespace("prompt.abstract", self.request_runtime_ctx, return_to = self)
        self.prompt_chat_history = RuntimeCtxNamespace("prompt.chat_history", self.request_runtime_ctx, return_to = self)
        self.prompt_input = RuntimeCtxNamespace("prompt.input", self.request_runtime_ctx, return_to = self)
        self.prompt_info = RuntimeCtxNamespace("prompt.info", self.request_runtime_ctx, return_to = self)
        self.prompt_instruct = RuntimeCtxNamespace("prompt.instruct", self.request_runtime_ctx, return_to = self)
        self.prompt_output = RuntimeCtxNamespace("prompt.output", self.request_runtime_ctx, return_to = self)
        self.prompt_files = RuntimeCtxNamespace("prompt.files", self.request_runtime_ctx, return_to = self)
        # Alias
        self.alias_manager = AliasManager(self)
        self._register_default_alias(self.alias_manager)

    def set_settings(self, settings_key: str, settings_value: any):
        self.settings.set(settings_key, settings_value)
        return self

    def set_request_prompt(self, key: str, value: any):
        self.request_runtime_ctx.set(f"prompt.{ key }", value)
        return self

    def get_request_prompt(self, key: str):
        return self.request_runtime_ctx.get(f"prompt.{ key }")

    def remove_request_prompt(self, key: str):
        self.request_runtime_ctx.remove(f"prompt.{ key }")
        return self

    def _register_default_alias(self, alias_manager):
        def set_model_settings(key: str, value: any, *, model_name:str = None):
            model_name = model_name if model_name else self.settings.get_trace_back("current_model")
            if model_name == None:
                raise Exception("[Model Settings] No model was appointed. Use .use_model(<model name>) or kwarg parameter model_name=<model_name> to set.")
            self.settings.update(f"model.{ model_name }.{ key }", value)

        alias_manager.register("use_model", lambda model_name: self.settings.set("current_model", model_name))
        alias_manager.register("set_model", set_model_settings)
        alias_manager.register("set_model_auth", lambda key, value, *, model_name=None: set_model_settings(f"auth.{ key }", value))
        alias_manager.register("set_model_url", lambda url, *, model_name=None: set_model_settings("url", url))
        alias_manager.register("set_model_option", lambda key, value, *, model_name=None: set_model_settings(f"options.{ key }", value))
        alias_manager.register("set_proxy", lambda proxy: self.settings.set("proxy", proxy))
        alias_manager.register("general", self.prompt_general.assign)
        alias_manager.register("role", self.prompt_role.assign)
        alias_manager.register("user_info", self.prompt_user_info.assign)
        alias_manager.register("abstract", self.prompt_abstract.assign)
        alias_manager.register("chat_history", self.prompt_chat_history.assign)
        alias_manager.register("input", self.prompt_input.assign)
        alias_manager.register("info", self.prompt_info.assign)
        alias_manager.register("instruct", self.prompt_instruct.assign)
        alias_manager.register("output", self.prompt_output.assign)
        alias_manager.register("file", self.prompt_files.append)
        alias_manager.register("files", self.prompt_files.extend)
        alias_manager.register("set_request_prompt", self.set_request_prompt)
        alias_manager.register("get_request_prompt", self.get_request_prompt)
        alias_manager.register("remove_request_prompt", self.remove_request_prompt) 
        
    async def get_event_generator(self, request_type: str=None):
        # Set Request Type
        self.request_runtime_ctx.set("request_type", request_type)
        # Erase response cache
        self.response_cache = {
            "prompt": {},
            "type": None,
            "reply": None,
        }
        # Confirm model name
        model_name = self.settings.get_trace_back("current_model")
        if not model_name:
            raise Exception(f"[Request] 'current_model' must be set. Use .use_model(<model_name>) to set.")
        # Load request plugin by model name
        request_plugin_instance = self.plugin_manager.get("request", model_name)(request = self)
        request_plugin_export = request_plugin_instance.export()
        # Generate request data
        if asyncio.iscoroutinefunction(request_plugin_export["generate_request_data"]):
            request_data = await request_plugin_export["generate_request_data"]()
        else:
            request_data = request_plugin_export["generate_request_data"]()
        if self.settings.get_trace_back("is_debug") == True:
            print("[Request Data]\n", json.dumps(request_data, indent=4, ensure_ascii=False))
        # Cache simple prompt
        self.response_cache["prompt"]["input"] = self.prompt_input.get()
        self.response_cache["prompt"]["output"] = self.prompt_output.get()
        if isinstance(self.response_cache["prompt"]["output"], (dict, list, set)):
            self.response_cache.update({"type": "JSON"})
        # Request and get response generator
        if asyncio.iscoroutinefunction(request_plugin_export["request_model"]):
            response_generator = await request_plugin_export["request_model"](request_data)
        else:
            response_generator = request_plugin_export["request_model"](request_data)
        # Broadcast response
        if asyncio.iscoroutinefunction(request_plugin_export["broadcast_response"]):
            broadcast_event_generator = await request_plugin_export["broadcast_response"](response_generator)
        else:
            broadcast_event_generator = request_plugin_export["broadcast_response"](response_generator)        
        self.response_cache["type"] = self.request_runtime_ctx.get("response:type")
        # Reset request runtime_ctx
        self.request_runtime_ctx.empty()
        return broadcast_event_generator

    async def get_result_async(self, request_type: str=None, *, return_generator:bool=False):
        try:
            is_debug = self.settings.get_trace_back("is_debug")
            event_generator = await self.get_event_generator(request_type)
            if is_debug:
                print("[Realtime Response]\n")
            def handle_response(response):
                if response["event"] == "response:delta":
                    self.response_generator.add(response["data"])
                    if is_debug:
                        print(response["data"], end="")
                if response["event"] == "response:done":
                    if is_debug:
                        print("\n--------------------------\n")
                        print("[Final Response]\n", response["data"], "\n--------------------------\n")
                    self.response_cache["reply"] = response["data"]
                    self.response_generator.end()

            if "__aiter__" in dir(event_generator):
                async for response in event_generator:
                    handle_response(response)
            else:
                for response in event_generator:
                    handle_response(response)
            
            if return_generator:
                self.request_runtime_ctx.empty()
                return self.response_generator.start()
            else:
                if (
                    isinstance(self.response_cache["prompt"]["output"], (dict, list, set))
                    or self.response_cache["type"] == "JSON"
                ):
                    temp_request = (
                        Request()
                            .set_settings("current_model", self.settings.get_trace_back("current_model"))
                            .set_settings("model", self.settings.get_trace_back("model"))
                    )
                    reply = await load_json(
                        self.response_cache["reply"],
                        self.response_cache["prompt"]["input"],
                        self.response_cache["prompt"]["output"],
                        temp_request,
                        is_debug = is_debug,
                    )
                    if isinstance(reply, str) and reply.startswith("$$$JSON_ERROR:"):
                        raise Exception(reply[14:])
                    self.response_cache["reply"] = json5.loads(reply) if isinstance(reply, str) else reply
                
                self.request_runtime_ctx.empty()
                return self.response_cache["reply"]
        except Exception as e:
            self.response_generator.end()
            self.request_runtime_ctx.empty()
            raise(e)

    def get_result(self, request_type: str=None, *, return_generator:bool=False):
        reply_queue = queue.Queue()
        is_debug = self.settings.get_trace_back("is_debug")
        def start_in_theard():
            asyncio.set_event_loop(asyncio.new_event_loop())
            loop = asyncio.get_event_loop()
            if is_debug:
                reply = loop.run_until_complete(self.get_result_async(request_type))
                reply_queue.put_nowait(reply)
            else:
                try:
                    reply = loop.run_until_complete(self.get_result_async(request_type))
                    reply_queue.put_nowait(reply)
                except Exception as e:
                    reply = None
                    raise Exception(f"[Request] Error: { str(e) }")
                finally:
                    loop.close()
        theard = threading.Thread(target=start_in_theard)
        theard.start()
        if return_generator:
            return self.response_generator.start()
        else:
            theard.join()        
            reply = reply_queue.get_nowait()
            return reply

    def start(self, request_type: str=None, *, return_generator=False):
        return self.get_result(request_type, return_generator=return_generator)

    def start_async(self, request_type: str=None):
        return self.get_result_async(request_type)


"""
helper method for lexer
"""


def is_ignore_token(c):
    """
    check if target character is ignore token
    """
    return c in "\t\n\v\f\r "


def match_stack(stack, tokens):
    """
    check if target stack match given tokens
    """
    pointer = len(stack)
    tokens_left = len(tokens)

    while True:
        tokens_left -= 1
        pointer -= 1
        if tokens_left < 0:
            break
        if pointer < 0:
            return False
        if stack[pointer] != tokens[tokens_left]:
            return False
    return True



import json5
from Agently.utils import Lexer
from .utils import ComponentABC
from Agently.utils import find_json

class Realtime(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.__is_enable = False
        self.__is_init = False
        self.__on_going_key_id = None
        self.__cached_value = {}
        self.__possible_keys = set()
        self.__emitted = set()
        self.__streaming_buffer = ""
        self.__realtime_value = None
    
    def use_realtime(self):
        self.__is_enable = True
        return self.agent
    
    def __scan_possible_keys(self, prompt_output_pointer, *, prefix:str=None):
        prefix = prefix if prefix != None else ""
        self.__possible_keys.add(prefix)
        if isinstance(prompt_output_pointer, dict):
            for key, value in prompt_output_pointer.items():
                self.__scan_possible_keys(value, prefix=prefix+f".{ key }")
        elif isinstance(prompt_output_pointer, list):
            for item in prompt_output_pointer:
                self.__scan_possible_keys(item, prefix=prefix+".[]")
        else:
            return
    
    async def __emit_realtime(self, key, indexes, delta, value):
        await self.agent.call_event_listeners(
            "realtime",
            {
                "key": key[1:],
                "indexes": indexes,
                "delta": delta,
                "value": value,
                "complete_value": self.__realtime_value, 
            }
        )

    async def __scan_realtime_value(self, key: str, indexes:list, value:any):
        indexes = indexes[:]
        key_id = (key, json5.dumps(indexes))
        if key_id in self.__emitted or key not in self.__possible_keys:
            return
        if isinstance(value, dict):
            for item_key, item_value in value.items():
                await self.__scan_realtime_value(key + f".{ item_key }", indexes, item_value)
            self.__cached_value[key_id] = value
        elif isinstance(value, list):
            for item_index, item_value in enumerate(value):
                temp = indexes[:]
                temp.append(item_index)
                await self.__scan_realtime_value(key + f".[]", temp, item_value)
            self.__cached_value[key_id] = value
        else:
            if isinstance(value, str):
                cached_value = self.__cached_value[key_id] if key_id in self.__cached_value else ""
                cached_value = cached_value if cached_value != None else ""
                delta = value.replace(cached_value, "")
                if len(delta) > 0:
                    await self.__emit_realtime(key, indexes, delta, value)
                    self.__cached_value[key_id] = value
                    self.__on_going_key_id = key_id
            else:
                self.__cached_value[key_id] = value
                self.__on_going_key_id = key_id
    def __judge_can_emit(self, current_key_id, on_going_key_id):
        if (not on_going_key_id[0].startswith(current_key_id[0])
            and on_going_key_id[0] + ".[]" != current_key_id[0]):
            return True
        current_indexes = json5.loads(current_key_id[1])
        on_going_indexes = json5.loads(on_going_key_id[1])
        for position, on_going_index in enumerate(on_going_indexes):
            if (len(current_indexes) >= position + 1
                and on_going_index > current_indexes[position]):
                return True
        return False
    
    async def __emit_waiting(self, *, is_done:bool=False):
        key_ids_to_del = []
        for key_id, value in self.__cached_value.items():
            indexes = json5.loads(key_id[1])
            if not is_done and key_id[0] == "":
                continue
            if key_id in self.__emitted:
                key_ids_to_del.append(key_id)
                continue
            if self.__on_going_key_id == None:
                continue
            if self.__judge_can_emit(key_id, self.__on_going_key_id) or is_done:
                if isinstance(value, dict):
                    await self.__emit_realtime(key_id[0], indexes, value, value)
                    self.__emitted.add(key_id)
                    key_ids_to_del.append(key_id)
                elif isinstance(value, list):
                    await self.__emit_realtime(key_id[0], indexes, value, value)
                    self.__emitted.add(key_id)
                    key_ids_to_del.append(key_id)
                elif isinstance(value, str):
                    await self.__emit_realtime(key_id[0] + ".$complete", indexes, value, value)
                    self.__emitted.add(key_id)
                    key_ids_to_del.append(key_id)
                else:
                    await self.__emit_realtime(key_id[0], indexes, value, value)
                    self.__emitted.add(key_id)
                    key_ids_to_del.append(key_id)
        for key_id in key_ids_to_del:
            del self.__cached_value[key_id]

    async def _suffix(self, event: str, data: any):
        if (
            not self.__is_enable
            or "type" not in self.agent.request.response_cache
            or self.agent.request.response_cache["type"] != "JSON"
        ):
            return None
        if not self.__is_init:
            prompt_output = self.agent.request.response_cache["prompt"]["output"]
            if isinstance(prompt_output, dict):
                self.__scan_possible_keys(prompt_output)
            self.__is_init = True
        if event == "response:delta":
            self.__streaming_buffer += data
            realtime_json_str = find_json(self.__streaming_buffer)
            if realtime_json_str != None:
                lexer = Lexer()
                lexer.append_string(realtime_json_str)
                try:
                    self.__realtime_value = json5.loads(lexer.complete_json())
                    await self.__scan_realtime_value("", [], self.__realtime_value)
                    await self.__emit_waiting()
                except ValueError:
                    return None
                except Exception as e:
                    raise(e)
        if event == "response:done":
            self.__streaming_buffer += data
            realtime_json_str = find_json(self.__streaming_buffer)
            if realtime_json_str != None:
                lexer = Lexer()
                lexer.append_string(realtime_json_str)
                try:
                    self.__realtime_value = json5.loads(lexer.complete_json())
                    await self.__scan_realtime_value("", [], self.__realtime_value)
                    await self.__emit_waiting(is_done=True)
                except ValueError:
                    return None
                except Exception as e:
                    raise(e)

    def export(self):
        return {
            "suffix": self._suffix,
            "alias": { "use_realtime": { "func": self.use_realtime } }
        }

def export():
    return ("Realtime", Realtime)


import uuid
import asyncio
import inspect
from .Runtime import RuntimeBranchState, RuntimeState, Snapshot, Checkpoint
from ..utils import RuntimeCtx
from .utils.exec_tree import disable_chunk_dep_ticket, create_new_chunk_slot_with_val
from .utils.find import find_by_attr
from .utils.runtime_supports import get_next_chunk_from_branch_queue
from .lib.BreakingHub import BreakingHub
from .lib.Store import Store
from .lib.constants import WORKFLOW_START_DATA_HANDLE_NAME, WORKFLOW_END_DATA_HANDLE_NAME, DEFAULT_INPUT_HANDLE_VALUE, DEFAULT_OUTPUT_HANDLE_VALUE, BUILT_IN_EXECUTOR_TYPES, EXECUTOR_TYPE_CONDITION

class MainExecutor:
    def __init__(self, workflow_id, settings: RuntimeCtx, logger, checkpoint: Checkpoint = None):
        # == Step 1. 初始化设定配置 ==
        self.workflow_id = workflow_id
        self.settings = settings
        self.max_execution_limit = self.settings.get('max_execution_limit') or 10
        self.logger = logger
        self.checkpoint = checkpoint
        # == Step 2. 初始化状态配置 ==
        # 中断器
        self.breaking_hub = BreakingHub(
            breaking_handler = self._handle_breaking,
            max_execution_limit=self.max_execution_limit
        )
        # 运行时状态
        self.runtime_state = RuntimeState(
            workflow_id=self.workflow_id,
            chunks_dep_state={},
            user_store=self.settings.get('store') or Store(),  # 运行时数据存储
            sys_store=self.settings.get('sys_store') or Store() # 运行时系统存储（如存储整个 workflow 的输入输出数据）
        )
        # 是否保留执行状态
        self.persist_state = self.settings.get('persist_state') == True
        self.persist_sys_state = self.settings.get('persist_sys_state') == True
        # 已注册的执行器
        self.registed_executors = {}
        # 执行节点字典
        self.chunks_map = {}

    async def start(self, executed_schema: dict, start_data: any = None, *, storage: dict = None):
        self.reset_all_runtime_status()
        # Set Initial Storage
        if storage and not isinstance(storage, dict):
            raise Exception(f"Initial storage can only be a dictionary.\nstorage = { storage }")
        elif storage and isinstance(storage, dict):
            self.runtime_state.user_store.update_by_dict(storage)

        # 初始化运行时状态
        self.runtime_state.update(chunks_dep_state=executed_schema.get('chunks_dep_state'))
        # 尝试灌入初始数据
        self.runtime_state.sys_store.set(WORKFLOW_START_DATA_HANDLE_NAME, start_data)
        self.chunks_map = executed_schema.get('chunk_map') or {}
        # 启动
        self.runtime_state.running_status = 'start'
        await self._execute_main(executed_schema.get('entries') or [])
        self.runtime_state.running_status = 'end'
        # 尝试返回执行结果
        return self.runtime_state.sys_store.get(WORKFLOW_END_DATA_HANDLE_NAME) or None
    
    async def start_from_snapshot(self, executed_schema: dict, snapshot: 'Snapshot'):
        # Step 1. 恢复各状态数据
        self.runtime_state.restore_from_snapshot(snapshot)
        self.chunks_map = executed_schema.get('chunk_map') or {}
        # Step 2. 恢复运行
        await self._execute_main(executed_schema.get('entries') or [])
        self.runtime_state.running_status = 'end'
        return self.runtime_state.sys_store.get(WORKFLOW_END_DATA_HANDLE_NAME) or None

    def regist_executor(self, name: str, executor):
        """
        注册执行器，传入执行器的名称及executor
        """
        self.registed_executors[name] = executor

    def unregist_executor(self, name: str):
        """
        取消注册执行器，传入执行器的名称
        """
        if name in self.registed_executors:
            del self.registed_executors[name]

        return self

    def reset_all_runtime_status(self):
        """重置状态配置"""
        self.runtime_state.running_status = 'idle'
        # 中断器
        self.breaking_hub = BreakingHub(
            breaking_handler=self._handle_breaking,
            max_execution_limit=self.max_execution_limit
        )
        # 运行时数据存储
        if not self.persist_state:
            self.runtime_state.user_store.remove_all()
        if not self.persist_sys_state:
            self.runtime_state.sys_store.remove_all()
        # 执行节点字典
        self.chunks_map = {}
    
    async def _execute_main(self, entries: list):
        """执行入口"""

        # 1、声明单个执行逻辑（异步）
        async def execute_from_entry(entry):
            branch_state = None
            # 恢复模式，直接使用已存储的状态
            if self.runtime_state.restore_mode:
                branch_state = self.runtime_state.get_branch_state(entry)
                # 已执行的分支直接跳过
                if branch_state.running_status in ['success', 'error']:
                    return
            # 正常模式，创建新分支状态
            else:
                branch_state = self.runtime_state.create_branch_state(entry)

            # 执行主逻辑
            branch_state.running_status = 'running'
            await self._execute_partial(entry, branch_state)
            branch_state.running_status = 'success'

        # 2、收集执行任务
        entry_tasks = [execute_from_entry(entry_chunk) for entry_chunk in entries]

        # 3、最后再统一执行
        await asyncio.gather(*entry_tasks)

    async def _execute_partial(self, chunk, branch_state: RuntimeBranchState):
        """分组执行的核心方法，整体基于队列的逻辑，确保整个过程可暂停/可恢复"""
        # 对于非恢复模式，或者恢复模式未执行态，默认初始化队列
        if not self.runtime_state.restore_mode or branch_state.get_chunk_status(chunk['id']) == 'idle':
            branch_state.running_queue.append(chunk['id'])

        while (branch_state.running_queue or branch_state.slow_queue) and branch_state.running_status != 'pause':
            # 本次是否为慢任务
            is_slow_task = (len(branch_state.running_queue) == 0) and (len(branch_state.slow_queue) > 0)
            current_chunk_id = get_next_chunk_from_branch_queue(branch_state)
            if not current_chunk_id:
                break

            chunk = self.chunks_map.get(current_chunk_id)
            if not chunk:
                raise SystemError(f'Target chunk({current_chunk_id}) not found')

            # 1、前置处理（如果是慢任务，先清空下游已执行的节点数据）
            if is_slow_task:
                self._chunks_clean_walker(chunk)

            # 2、执行 chunk
            has_been_executed = await self._execute_single_chunk(
                chunk=chunk,
                branch_state=branch_state,
                force_exec_loop=is_slow_task
            )
            # 清理，从队列中弹出本次执行的逻辑
            if is_slow_task:
                if len(branch_state.slow_queue):
                    branch_state.slow_queue.popleft()
            else:
                if len(branch_state.running_queue):
                    branch_state.running_queue.popleft()
            # 尝试自动保存 checkpoint
            if self.checkpoint and self.settings.get('auto_save_checkpoint') != False:
                await self.checkpoint.save_async(self.runtime_state)

            # 如果根本未执行过（如无执行票据），直接返回
            if not has_been_executed:
                continue

            # 3、将子任务放入执行队列中
            for next_info in chunk['next_chunks']:
                next_chunk = self.chunks_map.get(next_info['id'])
                if not next_chunk:
                    continue

                branch_state.running_queue.append(next_info['id'])

    async def _execute_single_chunk(self, chunk, branch_state: RuntimeBranchState, force_exec_loop = False):
        """执行完一个 chunk 自身（包含所有可用的依赖数据的组合）"""
        has_been_executed = False
        # 针对循环，不在本执行组内执行，存入缓执行组中，延后执行
        if (not force_exec_loop) and (chunk.get('loop_entry') == True) and (chunk['id'] in branch_state.visited_record):
            branch_state.slow_queue.append(chunk['id'])
            self.logger.debug(
                f"Put the loop starting chunk '{self._get_chunk_title(chunk)}' into the slow tasks queue for delayed execution")
            return has_been_executed

        # 获取执行chunk的依赖数据（每个手柄可能有多份就绪的数据）
        single_dep_map = self._extract_execution_single_dep_data(chunk)
        while (single_dep_map['is_ready'] and single_dep_map['has_ticket']):
            # 标识状态（仅需设置一次）
            if not has_been_executed:
                branch_state.update_chunk_status(chunk['id'], 'running')
            has_been_executed = True
            # 基于依赖数据快照，执行分组
            await self._execute_single_chunk_core(
                chunk=chunk,
                branch_state=branch_state,
                single_dep_map=single_dep_map['data']
            )
            # 执行后，收回本次用到的执行票据（仅在自身完了后才执行）
            self._disable_dep_execution_ticket(
                single_dep_map=single_dep_map['data'],
                chunk=chunk
            )
            # 再次更新获取依赖（如没有了，则停止了）
            single_dep_map = self._extract_execution_single_dep_data(chunk)
        # 更新结果状态
        if has_been_executed:
            branch_state.update_chunk_status(chunk['id'], 'success')
        return has_been_executed

    async def _execute_single_chunk_core(
        self,
        chunk,
        branch_state: RuntimeBranchState,
        single_dep_map: dict  # 依赖项（已拆组之后的）
    ):
        """根据某一份指定的的依赖数据，执行当前 chunk 自身（不包含下游 chunk 的执行调用）"""
        # 1、执行当前 chunk
        execute_id = str(uuid.uuid4())
        branch_state.executing_ids.append(execute_id)
        # self.logger.debug("With dependent data: ", single_dep_map)
        exec_res = await self._exec_chunk_with_dep_core(chunk, single_dep_map)
        exec_value = exec_res['value']
        condition_signal = exec_res['signal']

        self.breaking_hub.recoder(chunk)  # 更新中断器信息
        if chunk['id'] not in branch_state.visited_record:
            branch_state.visited_record.append(chunk['id'])  # 更新执行记录

        # 2、执行完成后，提取当前执行的结果，尝试将当前执行结果注入到下游的运行依赖插槽上 self.runtime_state.chunks_dep_state[][]['data_slots'][] = 'xxx'
        for next_info in chunk['next_chunks']:
            next_chunk = self.chunks_map.get(next_info['id'])
            if not next_chunk:
                continue

            # 针对目标 chunk 的目标 handle 的插槽位，注入最新的值（带执行票据的）。（注意如果是条件连接线，需要在条件满足时才更新）
            for next_rel_handle in next_info['handles']:
                source_handle = next_rel_handle['source_handle']
                target_handle = next_rel_handle['handle']
                source_value = None
                # 以下情况直接将完整值注入下游对应的插槽位置（此处的 target_handle）：执行结果为非 dict 类型，或者未定义上游 chunk 的输出句柄(此处的 source_handle)，或其输出句柄为默认全量输出句柄时
                if (not isinstance(exec_value, dict)) or (not source_handle) or (source_handle == DEFAULT_OUTPUT_HANDLE_VALUE):
                    source_value = exec_value
                else:
                    source_value = exec_value.get(source_handle)

                # 有条件的情况下，仅在条件满足时，才更新下游节点的数据
                condition_call = next_rel_handle.get('condition')
                if condition_call:
                    judge_res = condition_call(condition_signal, self.runtime_state.user_store)
                    connection_status = judge_res == True
                    self.logger.debug(
                        f"The connection status of '{self._get_chunk_title(chunk)}({source_handle})' to '{self._get_chunk_title(next_chunk)}({target_handle})': {connection_status}")
                    if not connection_status:
                        continue

                # 在下一个 chunk 的依赖定义中，找到与当前 chunk 的当前 handle 定义的部分，尝试更新其插槽值依赖
                self.logger.debug(
                    f"Try update chunk '{self._get_chunk_title(next_chunk)}' dep handle '{target_handle}' with value:{source_value}")
                deps = self.runtime_state.chunks_dep_state.get(next_chunk.get('id'))
                next_chunk_target_dep = find_by_attr(deps, 'handle', target_handle)
                if next_chunk_target_dep:
                    next_chunk_dep_slots = next_chunk_target_dep['data_slots'] or []
                    # 1、首先清空掉之前由当前节点设置，但票据已失效的值
                    next_chunk_target_dep['data_slots'] = next_chunk_dep_slots = [
                        slot for slot in next_chunk_dep_slots if not ((slot['updator'] == chunk['id']) and slot['execution_ticket'] == '')
                    ]

                    # 2、再把本次新的值加入到该下游 chunk 的对应输入点的插槽位中
                    next_chunk_dep_slots.append(
                        create_new_chunk_slot_with_val(chunk['id'], source_value))

        # 任务执行完后，清理执行中的状态
        branch_state.executing_ids.remove(execute_id)
    
    async def _exec_chunk_with_dep_core(self, chunk, specified_deps = {}):
        """ 执行任务（执行到此处的都是上游数据已就绪了的） """
        # 简化参数
        deps_dict = {}
        for dep_handle in specified_deps:
            deps_dict[dep_handle] = specified_deps[dep_handle]['value']
        
        input_value = deps_dict

        # 激进模式下的特殊处理
        if self.settings.get('mode') == 'aggressive':
            # 如果只有一个数据挂载，且为 default，则直接取出来作为默认值
            all_keys = list(deps_dict.keys())
            if len(all_keys) == 1 and all_keys[0] == DEFAULT_INPUT_HANDLE_VALUE:
                input_value = deps_dict['default']

        # 交给执行器执行
        executor_type = chunk['type']
        chunk_executor = self._get_chunk_executor(executor_type) or chunk.get('executor')
        # 是否内置的执行器（会追加系统信息）
        is_built_in_type = executor_type in BUILT_IN_EXECUTOR_TYPES
        exec_res = None
        if not chunk_executor:
            err_msg = f"Node {executor_type} Error-'{self._get_chunk_title(chunk)}'({chunk['id']}): The 'executor' is required but get 'NoneType'"
            self.logger.error(err_msg)
            # 主动中断执行
            raise Exception(err_msg)
        try:
            self.logger.info(f"Executing chunk '{self._get_chunk_title(chunk)}'")
            # 如果执行器是异步的，采用 await调用
            if inspect.iscoroutinefunction(chunk_executor):
                if is_built_in_type:
                    exec_res = await chunk_executor(input_value, self.runtime_state.user_store, sys_store=self.runtime_state.sys_store, chunk=chunk)
                else:
                    exec_res = await chunk_executor(input_value, self.runtime_state.user_store)
            else:
                if is_built_in_type:
                    exec_res = chunk_executor(input_value, self.runtime_state.user_store, sys_store=self.runtime_state.sys_store, chunk=chunk)
                else:
                    exec_res = chunk_executor(input_value, self.runtime_state.user_store)
        except Exception as e:
            self.logger.error(f"Node Execution Exception-'{self._get_chunk_title(chunk)}'({chunk['id']}):\n {e}")
            # 处理 checkpoint 自动保存到默认的 checkpoint 点
            if self.checkpoint and self.settings.get('save_checkpoint_on_error') != False:
                await self.checkpoint.save_async(self.runtime_state)
                self.logger.info("Checkpoint has been automatically saved.")
            # 主动中断执行
            raise Exception(e)

        exec_value = exec_res
        condition_signal = None
        # 条件节点，拆分为条件信号和执行结果两部分
        if chunk['type'] == EXECUTOR_TYPE_CONDITION:
            exec_value = exec_res.get('values')
            condition_signal = exec_res.get('condition_signal')
        return {
            "value": exec_value,
            "signal": condition_signal
        }
    
    def _extract_execution_single_dep_data(self, chunk):
        """实时获取某个 chunk 的一组全量可执行数据（如没有，则返回 None）"""
        deps = self.runtime_state.chunks_dep_state.get(chunk.get('id'))
        if not deps or len(deps) == 0:
            return {"is_ready": True, "data": None, "has_ticket": True}
        single_dep_map = {}
        exist_exec_ticket = False

        for dep in deps:
            slots = dep['data_slots'] or []
            handle_name = dep['handle']
            # 暂存的就绪的数据（循环中会不停更新，后头的会覆盖前头的），注意，就绪的数据也有可能就是 None，所以 None 不代表没有就绪数据
            tmp_ready_slot = None
            # 是否有就绪的值
            has_ready_value = False

            for slot in slots:
                # 找到就绪的数据
                if slot['is_ready']:
                    has_ready_value = True
                    # 先暂存就绪的数据作为临时数据（后面的会覆盖前头的）
                    tmp_ready_slot = slot

                    # 如果目前全局还没遇到过有票据，且当前为有票据的情况，则作为本次执行的消耗票据
                    if slot['execution_ticket'] and not exist_exec_ticket:
                        exist_exec_ticket = True
                        single_dep_map[handle_name] = slot
                        break

            # 如果跑完所有，还是没设置值（可能没有就绪的数据，或者就绪的数据都是有票据的），需要从就绪的数据中强行取一个作为本次的值
            if (handle_name not in single_dep_map) and has_ready_value:
                single_dep_map[handle_name] = tmp_ready_slot

            # 如果本轮跑完都没有设置值，则标识该 handle 数据未就绪，直接返回
            if handle_name not in single_dep_map:
                return {"is_ready": False, "data": None, "has_ticket": exist_exec_ticket}

        return {"is_ready": True, "data": single_dep_map, "has_ticket": exist_exec_ticket}

    def _disable_dep_execution_ticket(self, single_dep_map, chunk):
        """销毁chunk对应的依赖执行票据（一般在执行结束后操作）"""
        for dep_handle in single_dep_map:
            # 找到被执行的 id
            effect_id = single_dep_map[dep_handle]['id']
            deps = self.runtime_state.chunks_dep_state.get(chunk['id'])
            for dep in deps:
                slots = dep['data_slots'] or []
                for slot in slots:
                    # 将当前 chunk 中对应手柄的数据项中的对应目标数据的执行票据收回
                    if slot['id'] == effect_id:
                        disable_chunk_dep_ticket(slot)

    def _get_chunk_executor(self, name: str):
        """ 根据类型名称获取执行器 """
        return self.registed_executors.get(name)

    def _chunks_clean_walker(self, root_chunk):
        """尝试对某个节点以下的分支做一轮清理工作"""

        visited_record = []
        def clean_core(chunk):
            for next_info in chunk['next_chunks']:
                next_chunk = self.chunks_map.get(next_info['id'])
                if not next_chunk:
                    continue

                # 同一个发起方的清理，只执行一次，避免死循环
                visited_symbol = f"{chunk['id']}-2-{next_chunk['id']}"
                if (visited_symbol in visited_record) or (next_chunk['id'] == root_chunk['id']):
                    continue

                visited_record.append(visited_symbol)
                effect_handles = [handle_desc['handle'] for handle_desc in next_info['handles']]
                deps = self.runtime_state.chunks_dep_state.get(next_chunk.get('id'))
                for dep in deps:
                    data_slots = dep['data_slots']
                    if len(data_slots):
                        for i in range(len(data_slots) - 1, -1, -1):
                            data_slot = data_slots[i]
                            # 如果当前的输入句柄是受上游影响的，则清理掉
                            if (dep['handle'] in effect_handles) and (data_slot['updator'] == '' or data_slot['updator'] == chunk['id']):
                                del data_slots[i]
                clean_core(next_chunk)

        clean_core(root_chunk)
    
    def _get_chunk_title(self, chunk):
        return chunk["title"] or f'chunk-{chunk["id"]}' or 'Unknow chunk'

    def _handle_breaking(self, chunk, type):
        """处理中断"""
        self.logger.error(
            f"Exceeded maximum execution limit: {self._get_chunk_title(chunk)}")
        # 中断之前处理相关逻辑
        # 主动中断执行
        raise Exception(
            f"Exceeded maximum execution limit: {self._get_chunk_title(chunk)}")



import types
import asyncio
from .utils import ComponentABC
from Agently.utils import RuntimeCtx, RuntimeCtxNamespace

class EventListener(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.agent_listeners = RuntimeCtxNamespace("event_listeners", self.agent.agent_runtime_ctx)
        self.listeners = RuntimeCtxNamespace("event_listeners", RuntimeCtx(parent=self.agent.agent_runtime_ctx))
        self.listeners.set({})
        self.async_tasks = []

    def add(self, event:str, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        event = event.replace(".", "->")
        if event.startswith("realtime:"):
            event_data = event.replace(" ", "").split(":")
            hooks = event_data[1].replace("->", ".").split("&")
            hook_list = []
            for hook in hooks:
                hook_key_indexes = hook.split("?")
                hook_key = hook_key_indexes[0]
                hook_indexes = None
                if len(hook_key_indexes) > 1:
                    hook_indexes = hook_key_indexes[1].split(",")
                    for index, item in enumerate(hook_indexes):
                        if item.startswith("(") and item.endswith(")"):
                            items_in_item = item[1:-1].split("|")
                        else:
                            items_in_item = [item]
                        for i, value in enumerate(items_in_item):    
                            try:
                                items_in_item[i] = int(value)
                            except:
                                items_in_item[i] = None
                        hook_indexes[index] = items_in_item
                hook_list.append((hook_key, hook_indexes))
                async def realtime_hook_handler(data):
                    for hook in hook_list:
                        hook_key = hook[0]
                        hook_indexes = hook[1]
                        if (data["key"] == hook_key):
                            if hook_indexes and len(hook_indexes) <= len(data["indexes"]):
                                can_call = True
                                for position, hook_index in enumerate(hook_indexes):
                                    if hook_index != None and data["indexes"][position] not in hook_index:
                                        can_call = False
                                if can_call:
                                    if asyncio.iscoroutinefunction(listener):
                                        await listener(data)
                                    else:
                                        listener(data)
                            else:
                                if asyncio.iscoroutinefunction(listener):
                                    await listener(data)
                                else:
                                    listener(data)
                if event not in (self.listeners.get(trace_back=False) or {}):
                    self.listeners.update(event, [])
                self.listeners.append("realtime", { "listener": realtime_hook_handler, "is_await": is_await })
        else:
            if is_agent_event:
                if event not in (self.agent_listeners.get(trace_back=False) or {}):
                    self.agent_listeners.update(event, [])
                self.agent_listeners.append(event, { "listener": listener, "is_await": is_await })
                return self.agent
            else:
                if event not in (self.listeners.get(trace_back=False) or {}):
                    self.listeners.update(event, [])
                self.listeners.append(event, { "listener": listener, "is_await": is_await })
                return self.agent

    def on_delta(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("response:delta", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent

    def on_done(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("response:done", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent

    def on_finally(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("response:finally", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent

    def on_realtime(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("realtime", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent
    
    async def call_event_listeners(self, event: str, data: any):
        event = event.replace(".", "->")
        listeners = self.listeners.get_trace_back() or {}
        if event in listeners:
            for listener_info in listeners[event]:
                if asyncio.iscoroutinefunction(listener_info["listener"]):
                    if listener_info["is_await"]:
                        await listener_info["listener"](data)
                    else:
                        self.async_tasks.append(asyncio.create_task(listener_info["listener"](data)))
                else:
                    listener_info["listener"](data)
        return self.agent

    async def _suffix(self, event:str, data: any):
        await self.call_event_listeners(event, data)
        if event == "response:finally":
            for async_task in self.async_tasks:
                await async_task
            self.async_tasks = []
            self.listeners.set({})

    def export(self):
        return {
            "prefix": None,
            "suffix": self._suffix,
            "alias": {
                "add_event_listener": { "func": self.add },
                "on_delta": { "func": self.on_delta },
                "on_done": { "func": self.on_done },
                "on_finally": { "func": self.on_finally },
                "on_realtime": { "func": self.on_realtime },
                "call_event_listeners": { "func": self.call_event_listeners },
            },
        }

def export():
    return ("EventListener", EventListener)


def validate_dict(input_dict, required_keys = []):
    # 检查是否包含所有必需的键
    for key in required_keys:
        if key not in input_dict:
            return {'status': False, 'key': key }
    
    return { 'status': True }


# 基于 MiniMax Pro 接口：https://www.minimaxi.com/document/guides/chat-model/pro/api
from .utils import RequestABC, to_prompt_structure, to_instruction, to_json_desc, find_json
from Agently.utils import RuntimeCtxNamespace
import httpx
import json

DEFAULT_AI_BOT_NAME = "智能助理"
DEFAULT_USER_NAME = "用户"
# 和身份相关的字段
REL_ROLE_KEY_NAMES = ['角色', '身份', 'role', '姓名', '名称', 'name', '扮演', 'role-playing']

class MiniMax(RequestABC):
    def __init__(self, request):
        self.request = request
        self.request_type = self.request.request_runtime_ctx.get("request_type", "chat")
        if self.request_type == None:
            self.request_type = "chat"

        self.model_name = "MiniMax"
        self.model_settings = RuntimeCtxNamespace(
            f"model.{self.model_name}", self.request.settings)
        options = self.model_settings.get_trace_back("options", {})
        self.model_version = options.get('model') or "abab5.5-chat"

        # abab5.5 模型的情况，启用 func 辅助 json 格式输出（保障结构稳定性）
        self.use_func_fix_json = self.model_version.startswith('abab5.5')

    def generate_scene_info(self):
        """标准化业务场景相关的字段"""
        req_messages = []
        req_bot_desc = []

        # 处理角色信息（包含角色名称提取和角色信息设定）
        bot_role_name = DEFAULT_AI_BOT_NAME
        role_data = self.request.request_runtime_ctx.get_trace_back(
            "prompt.role")
        if role_data:
            role_data = fix_define_to_obj(role_data, '角色信息')
            for role_key in REL_ROLE_KEY_NAMES:
                role_val = get_value_ignore_case(role_data, role_key)
                if role_val:
                    bot_role_name = role_val
                    break
            req_bot_desc.append(to_instruction(role_data))

        # 处理指令
        general_instruction_data = self.request.request_runtime_ctx.get_trace_back(
            "prompt.general")
        if general_instruction_data:
            req_bot_desc.append(
                f"[重要指导说明]{to_instruction(general_instruction_data)}")

        # 处理用户信息
        user_role_name = DEFAULT_USER_NAME
        user_info_data = self.request.request_runtime_ctx.get_trace_back(
            "prompt.user_info")
        if user_info_data:
            user_info_data = fix_define_to_obj(user_info_data, '用户信息')
            for rel_key in REL_ROLE_KEY_NAMES:
                user_val = get_value_ignore_case(user_info_data, rel_key)
                if user_val:
                    user_role_name = user_val
                    break
            req_messages.append(to_user_msg(
                f"[这是我的信息介绍]：{to_instruction(user_info_data)}", user_role_name))
            req_messages.append(to_bot_msg('收到', bot_role_name))

        # 主题摘要
        headline_data = self.request.request_runtime_ctx.get_trace_back(
            "prompt.abstract")
        if headline_data:
            req_messages.append(to_user_msg(
                f"[这是我们要聊的话题]：{to_instruction(headline_data)}", user_role_name))
            req_messages.append(to_bot_msg('收到', bot_role_name))

        # 历史信息
        chat_history_data = self.request.request_runtime_ctx.get(
            "prompt.chat_history")
        if chat_history_data:
            req_messages.extend(fix_history(
                chat_history_data,
                user_name=user_role_name,
                robot_name=bot_role_name
            ))

        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get(
            "prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get(
            "prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get(
            "prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get(
            "prompt.output")
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception(
                "[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")

        func_args = {}
        prompt_dict = {}
        if prompt_info_data:
            prompt_dict["[补充信息]"] = str(prompt_info_data)
        if prompt_instruct_data:
            prompt_dict["[处理规则]"] = to_instruction(prompt_instruct_data)

        if prompt_output_data:
            if isinstance(prompt_output_data, (dict, list, set)):
                # 采用 func 辅助修正格式
                if self.use_func_fix_json:
                    func_args["functions"] = [{
                        "name": "output_result",
                        "description": "输出结果",
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "result": {
                                    "type": "object",
                                    "description": to_json_desc(prompt_output_data)
                                }
                            },
                            "required": [ "result" ]
                        }
                    }]
                    func_args['function_call'] = {
                        "type": "specific",
                        "name": "output_result"
                    }
                # 常规场景
                else:
                    prompt_dict["[输出要求]"] = {
                        "TYPE": "可被解析的JSON字符串",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }

                self.request.request_runtime_ctx.set("response:type", "JSON")
            else:
                prompt_dict["[输出要求]"] = str(prompt_output_data)

        if len(prompt_dict) > 0:
            req_bot_desc.append(to_prompt_structure(prompt_dict))

        # 处理输入
        req_messages.append(to_user_msg(
            to_instruction(prompt_input_data), user_role_name))

        # 兜底 prompt
        if len(req_bot_desc) == 0:
            req_bot_desc.append(f"尽可能准确且专业地回答用户问题")
        
        return {
            **func_args,
            "bot_setting": [{
                "bot_name": bot_role_name,
                "content": '\n----\n'.join(req_bot_desc)
            }],
            "reply_constraints": {
                "sender_type": "BOT",
                "sender_name": bot_role_name
            },
            "messages": req_messages
        }

    def generate_request_data(self):
        options = self.model_settings.get_trace_back("options", {})
        scene_info = self.generate_scene_info()

        return {
            **scene_info,
            **options,
            "model": self.model_version,
            "stream": True
        }

    async def request_model(self, request_data: dict):
        group_id = self.model_settings.get_trace_back("auth.group_id")
        api_key = self.model_settings.get_trace_back("auth.api_key")

        if not group_id or not api_key:
            raise Exception(
                "[Request] Missing 'auth.group_id', 'auth.api_key' in model_settings. At least set value to one of them.")

        proxy = self.request.settings.get_trace_back("proxy")
        client_params = {}
        if proxy:
            client_params["proxy"] = proxy

        async with httpx.AsyncClient(**client_params) as client:
            async with client.stream(
                "POST",
                f"https://api.minimax.chat/v1/text/chatcompletion_pro?GroupId={group_id}",
                headers={
                    "Authorization": f"Bearer {api_key}",
                    "Content-Type": "application/json"
                },
                data=json.dumps(request_data).encode()
            ) as response:
                async for chunk in response.aiter_lines():
                    yield chunk

    async def broadcast_response(self, response_generator):
        full_res = ''
        full_res_raw = {}
        delta_content = ''
        old_delta_content = ''
        async for part in response_generator:
            origin_part = part[6:] if part.startswith('data:') else part
            if origin_part:
                chunk = json.loads(origin_part)
                # 异常判断
                if chunk.get('base_resp') and chunk.get('base_resp').get('status_code'):
                    raise Exception(f"[Response Error]{json.dumps(chunk.get('base_resp'))}")

                content_chunk = chunk['choices'][0]
                content_chunk_message = content_chunk['messages'][0]
                function_call_chunk = content_chunk_message.get('function_call')

                # func 辅助的场景
                if self.use_func_fix_json and function_call_chunk:
                    delta_content = extract_res_from_func_call(
                        function_call_chunk.get('arguments')
                    )
                    
                    if delta_content:
                        # 此处是直接替换的
                        full_res = delta_content
                        # 尝试计算出增量
                        delta_content = full_res[len(old_delta_content):]
                        old_delta_content = full_res

                    # 计算结束
                    if content_chunk.get('finish_reason') == 'stop':
                        full_res_raw = chunk
                        full_chunk_content = extract_res_from_func_call(
                            (content_chunk_message.get(
                                'function_call') or {}).get('arguments')
                        )
                    else:
                        yield ({"event": "response:delta_origin", "data": chunk})
                        yield ({"event": "response:delta", "data": delta_content})
                # 常规场景
                else:
                    full_res_raw = chunk
                    full_chunk_content = content_chunk['messages'] or []
                    delta_content = full_chunk_content[0]['text']
                    
                    if content_chunk.get('finish_reason') != 'stop':
                        full_res += delta_content
                        yield ({"event": "response:delta_origin", "data": chunk})
                        yield ({"event": "response:delta", "data": delta_content})

        yield ({"event": "response:done_origin", "data": full_res_raw})
        yield ({"event": "response:done", "data": full_res})

    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }


# 修复定义信息
def fix_define_to_obj(value, default_key = '信息'):
    if isinstance(value, str):
        return {
            default_key: value
        }
    return value

# 忽略大小写，从 dict 中获取对应键的值
def get_value_ignore_case(dictionary, key):
    for k, v in dictionary.items():
        if k.lower() == key.lower():
            return v
    return None

# 用户消息体构建
def to_user_msg(msg, user_name):
    return {"sender_type": "USER", "sender_name": user_name, "text": msg}

# 回复消息体构建
def to_bot_msg(msg, bot_name):
    return {"sender_type": "BOT", "sender_name": bot_name, "text": msg}

# 将标准历史消息转成 MiniMax 适配的格式
def fix_history(chat_history, user_name = '', robot_name = ''):
    def fix(chat_item):
        return {
            "sender_type": 'USER' if chat_item['role'] == 'user' else 'BOT',
            "sender_name": user_name if chat_item['role'] == 'user' else robot_name,
            "text": chat_item['content']
        }
    return map(fix, chat_history)

# 从 function call 中获取参数数据
def extract_res_from_func_call(args_str: str):
    return (args_str or '').replace("result=", "", 1)

# 主导出函数
def export():
    return "MiniMax", MiniMax



from ..utils import PluginManager, RuntimeCtx, RuntimeCtxNamespace
from .._global import global_plugin_manager, global_storage, global_settings

class FacilityManager(object):
    def __init__(self, *, storage: object=global_storage, parent_plugin_manager: object=global_plugin_manager, parent_settings: object = global_settings):
        # init plugin manager
        self.plugin_manager = PluginManager(parent = parent_plugin_manager)
        # use global storage
        self.storage = storage
        # init facility settings
        self.settings = RuntimeCtx(parent = global_settings)
        # install facilities
        self.refresh_plugins()

    def refresh_plugins(self):
        facilities = self.plugin_manager.get("facility")
        for facility_name, FacilityPluginClass in facilities.items():
            setattr(self, facility_name, FacilityPluginClass(storage = self.storage, plugin_manager = self.plugin_manager, settings = self.settings))

    def set_settings(self, settings_key: str, settings_value: any):
        self.settings.set(settings_key, settings_value)
        return self

    def list(self):
        result = {}
        facility_names = [\
            attr_name for attr_name in dir(self)\
            if not attr_name.startswith("_")\
            and attr_name not in ("plugin_manager", "storage", "list")\
            and isinstance(getattr(self, attr_name), object)\
        ]
        for facility_name in facility_names:
            facility = getattr(self, facility_name)
            facility_method_names = [\
                method_name for method_name in dir(facility)\
                if not method_name.startswith("_")\
                and callable(getattr(facility, method_name))\
            ]
            result[facility_name] = facility_method_names
        return result


from typing import List
from Agently.utils import StorageDelegate
from Agently._global import global_plugin_manager, global_settings

class CheckpointRepository:
  def __init__(self, storage = None) -> None:
    self.storage = storage or StorageDelegate(
      db_name="checkpoint",
      plugin_manager=global_plugin_manager,
      settings=global_settings
    )

  async def save(self, checkpoint_id, name: str, data: any):
    self.storage.set(table_name=checkpoint_id, key=name, value=data)
    return self

  async def remove(self, checkpoint_id, name: str):
    self.storage.remove(table_name=checkpoint_id, key=name)
    return self

  async def get(self, checkpoint_id, name: str):
    return self.storage.get(checkpoint_id, key=name)

  async def get_all(self, checkpoint_id, names: List[str] = None):
    return self.storage.get_all(checkpoint_id, keys=names)



import threading
from ..utils import DataGenerator, DataOps
from .buildin import gradio_app, streamlit_app, shell_app

class AppConnector(object):
    def __init__(
            self,
            *,
            app_name:str="gradio",
            message_handler:callable=None,
            binded_agent:object=None,
        ):
        self.app_name = app_name
        self.message_handler = message_handler
        self.message_handler_thread = None
        self.binded_agent = binded_agent
        self.buffer = ""
        self.app_dict = {
            "gradio": gradio_app,
            "streamlit": streamlit_app,
            "shell": shell_app,  
        }
        self.runtime = DataOps()
        self.data_generator = DataGenerator()
    
    def register_app(self, type_name:str, app:callable):
        self.app_dict.update({ type_name: app })
        return self
    
    def use_app(self, app_name:str):
        self.app_name = app_name
        return self
    
    def set_message_handler(self, message_handler: callable):
        self.message_handler = message_handler
        return self
    
    def run_message_handler(self, *args, **kwargs):
        self.message_handler_thread = threading.Thread(
            target=self.message_handler,
            args=args,
            kwargs=kwargs,
        )
        self.message_handler_thread.start()
        return self
    
    def bind_agent(self, agent):
        self.binded_agent = agent
        if self.message_handler == None:
            self.set_message_handler(
                lambda message, chat_history: agent.chat_history(chat_history).input(message).start()
            )
        return self
    
    def refresh_binded_agent(self):
        if self.binded_agent:
            @self.binded_agent.on_event("delta")
            def delta_handler(data):
                self.emit_delta(data)
            @self.binded_agent.on_event("done")
            def done_handler(data):
                self.emit_done(data)
        return self
    
    def emit(self, event: str, data: any):
        self.data_generator.event.emit(event, data)
        return self

    def emit_delta(self, delta:str=None):
        self.buffer += delta if delta else ""
        self.emit("delta", delta)
        self.emit("buffer", self.buffer)
        return self
    
    def emit_buffer(self, buffer:str=None):
        self.buffer = buffer if buffer else ""
        self.emit("buffer", self.buffer)
        return self
    
    def emit_done(self, final_result:str=None):
        self.emit("done", final_result if final_result else self.buffer)
        self.buffer = ""
        return self

    def on(self, event:str, handler:callable):
        self.data_generator.event.on(event, handler)
        return self
    
    def reset_data_generator(self):
        self.data_generator = DataGenerator()
        return self

    def run(self, app_name:str=None, **kwargs):
        app_name = app_name or self.app_name
        self.app_dict[app_name](self, **kwargs)
        if self.message_handler_thread:
            self.message_handler_thread.join()
            self.message_handler_thread = None


import copy

class NamespaceOps(object):
    def __init__(self, namespace_name: str, data_ops: object, *, return_to: object=None):
        self.namespace_name = namespace_name
        self.data_ops = data_ops
        self.return_to = return_to if return_to else self

    def assign(self, input:any, desc: any=None):
        if input and desc == None:
            if isinstance(input, dict):
                return self.update(input)
            elif isinstance(input, list):
                return self.extend(input)
            else:
                return self.set(input)
        if input and desc != None:
            current_content = self.get(input)
            if isinstance(current_content, list):
                if isinstance(desc, list):
                    return self.extend(input, desc)
                else:
                    return self.append(input, desc)
            elif isinstance(desc, dict):
                return self.update(input, desc)
            else:
                return self.set(input, desc)


    def set(self, keys_with_dots: any, value: any=None):
        if value == None:
            self.data_ops.set(self.namespace_name, keys_with_dots)
        else:
            self.data_ops.set(f"{ self.namespace_name }.{ keys_with_dots }", value)
        return self.return_to

    def delta(self, keys_with_dots: str, value: any):
        if isinstance(value, dict):
            for k, v in value.items():
                self.delta(f"{ keys_with_dots }.{ k }", v)
        else:
            self.append(keys_with_dots, value)

    def append(self, keys_with_dots: any, value: any=None):
        if value == None:
            self.data_ops.append(self.namespace_name, keys_with_dots)
        else:
            self.data_ops.append(f"{ self.namespace_name }.{ keys_with_dots }", value)
        return self.return_to

    def extend(self, keys_with_dots: any, value: any=None):
        if value == None:
            self.data_ops.extend(self.namespace_name, keys_with_dots)
        else:
            self.data_ops.extend(f"{ self.namespace_name }.{ keys_with_dots }", value)
        return self.return_to

    def update(self, keys_with_dots: any, value: any=None):
        if value == None:
            self.data_ops.update(self.namespace_name, keys_with_dots)
        else:
            self.data_ops.update(f"{ self.namespace_name }.{ keys_with_dots }", value)
        return self.return_to
        
    def get(self, keys_with_dots: str = None, default = None, *, no_copy: bool = False):
        return self.data_ops.get(f"{ self.namespace_name }.{ keys_with_dots }" if keys_with_dots else self.namespace_name, default, no_copy = no_copy)

    def remove(self, keys_with_dots: str):
        return self.data_ops.remove(f"{ self.namespace_name }.{ keys_with_dots }")

    def empty(self):
        return self.data_ops.remove(self.namespace_name)

class DataOps(object):
    def __init__(self, *, target_data:dict=None, no_copy: bool=False):
        self.target_data = target_data or {}
        self.no_copy = no_copy

    def __locate_pointer(self, keys_with_dots: str):
        keys = keys_with_dots.split('.')
        pointer = self.target_data
        current_key = None
        for key in keys:
            if current_key:
                pointer = pointer[current_key]
            current_key = key
            if key not in pointer:
                pointer[key] = {}
        return pointer, current_key

    def set(self, keys_with_dots: str, value: any):
        pointer, key = self.__locate_pointer(keys_with_dots)
        pointer[key] = value
        return self

    def delta(self, keys_with_dots: str, value: any):
        if isinstance(value, dict):
            for k, v in value.items():
                self.delta(f"{ keys_with_dots }.{ k }", v)
        else:
            if self.get(keys_with_dots):
                self.append(keys_with_dots, value)
            else:
                self.set(keys_with_dots, value)

    def append(self, keys_with_dots: str, value: any):
        pointer, key = self.__locate_pointer(keys_with_dots)
        if isinstance(pointer[key], list):
            pointer[key].append(value)
        elif pointer[key] == {}:
            pointer[key] = []
            pointer[key].append(value)
        else:
            pointer[key] = [pointer[key]]
            pointer[key].append(value)
        return self

    def extend(self, keys_with_dots: str, value: any):
        if not isinstance(value, list):
            value = [value]
        pointer, key = self.__locate_pointer(keys_with_dots)
        if isinstance(pointer[key], list):
            pointer[key].extend(value)
        elif pointer[key] == {}:
            pointer[key] = []
            pointer[key].extend(value)
        else:
            pointer[key] = [pointer[key]]
            pointer[key].extend(value)
        return self 

    def __update_dict(self, pointer: dict, pointer_key: str, dict_item: any):
        if isinstance(dict_item, dict):
            for key in dict_item.keys():
                if key not in pointer[pointer_key]:
                    pointer[pointer_key][key] = {}
                if isinstance(dict_item[key], dict):
                    self.__update_dict(pointer[pointer_key], key, dict_item[key])
                else:
                    if dict_item[key] != None:
                        pointer[pointer_key][key] = dict_item[key]
        else:
            if dict_item != None:
                pointer[pointer_key] = dict_item

    def update(self, keys_with_dots: str, value: any):
        pointer, key = self.__locate_pointer(keys_with_dots)
        self.__update_dict(pointer, key, value)
        return self

    def update_by_dict(self, data_dict: dict):
        for key, value in data_dict.items():
            self.update(key, value)
        return self

    def _deep_get(self, data):
        result = None
        if isinstance(data, dict):
            result = {}
            for key, value in data.items():
                result.update({ key: self._deep_get(data[key]) })
            return result
        elif isinstance(data, list):
            result = []
            for item in data:
                result.append(self._deep_get(item))
            return result
        elif isinstance(data, set):
            result = set()
            for item in list(data):
                result.add(self._deep_get(item))
            return result
        elif isinstance(data, tuple):
            result = []
            for item in list(data):
                result.append(self._deep_get(item))
            return tuple(result)
        elif hasattr(data, '__class__') and hasattr(data, '__module__') and data.__module__ != 'builtins':
            return data
        else:
            return copy.deepcopy(data)

    def get(self, keys_with_dots: str = None, default: str=None, *, no_copy: bool = False):
        if keys_with_dots:
            keys = keys_with_dots.split('.')
            pointer = self.target_data
            for key in keys:
                if key not in pointer:
                    return default
                else:
                    pointer = pointer[key]
            if self.no_copy or no_copy:
                return pointer
            else:
                if hasattr(pointer, '__class__') and hasattr(pointer, '__module__') and pointer.__module__ != 'builtins':
                    return pointer
                else:
                    try:
                        return copy.deepcopy(pointer)
                    except TypeError:
                        return self._deep_get(pointer)
        else:
            if self.no_copy or no_copy:
                return self.target_data
            else:
                if hasattr(self.target_data, '__class__') and hasattr(self.target_data, '__module__') and self.target_data.__module__ != 'builtins':
                    return self.target_data
                else:
                    try:
                        return copy.deepcopy(self.target_data)
                    except TypeError:
                        return self.target_data

    def remove(self, keys_with_dots: str):
        pointer, key = self.__locate_pointer(keys_with_dots)
        del pointer[key]
        return self

    def empty(self):
        self.target_data = {}
        return self


from .StartExecutor import start_executor
from .EndExecutor import end_executor
from .ConditionExecutor import condition_executor
from ...MainExecutor import MainExecutor
from ...lib.constants import EXECUTOR_TYPE_START, EXECUTOR_TYPE_END, EXECUTOR_TYPE_CONDITION

def mount_built_in_executors(main_executor: MainExecutor):
    """
    挂载内置的执行器
    """
    main_executor.regist_executor(EXECUTOR_TYPE_START, start_executor)
    main_executor.regist_executor(EXECUTOR_TYPE_END, end_executor)
    main_executor.regist_executor(EXECUTOR_TYPE_CONDITION, condition_executor)



from .utils import RequestABC, to_prompt_structure, to_instruction, to_json_desc, format_request_messages
from Agently.utils import RuntimeCtxNamespace
import erniebot


class Ernie(RequestABC):
    def __init__(self, request):
        self.request = request
        self.request_type = self.request.request_runtime_ctx.get("request_type", "chat")
        if self.request_type == None:
            self.request_type = "chat"
        self.model_name = "ERNIE"
        self.model_settings = RuntimeCtxNamespace(f"model.{self.model_name}", self.request.settings)
        if not self.model_settings.get_trace_back("message_rules.no_multi_system_messages"):
            self.model_settings.set("message_rules.no_multi_system_messages", True)
        if not self.model_settings.get_trace_back("message_rules.strict_orders"):
            self.model_settings.set("message_rules.strict_orders", True)
        if not self.model_settings.get_trace_back("message_rules.no_multi_type_messages"):
            self.model_settings.set("message_rules.no_multi_type_messages", True)    

    def _create_client(self):
        if self.request_type == "chat":
            client = erniebot.ChatCompletion
        elif self.request_type == "embedding":
            client = erniebot.Embedding
        elif self.request_type == "image":
            client = erniebot.Image
        return client

    def construct_request_messages(self):
        # init request messages
        request_messages = []
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get(
            "prompt.general")
        if general_instruction_data:
            request_messages.append({"role": "system",
                                     "content": f"[重要指导说明]\n{to_instruction(general_instruction_data)}"})
        # - role
        role_data = self.request.request_runtime_ctx.get("prompt.role")
        if role_data:
            request_messages.append(
                {"role": "system", "content": f"[角色及行为设定]\n{to_instruction(role_data)}"})
        # - user info
        user_info_data = self.request.request_runtime_ctx.get("prompt.user_info")
        if user_info_data:
            request_messages.append(
                {"role": "system", "content": f"[用户信息]\n{to_instruction(user_info_data)}"})
        # - abstract
        headline_data = self.request.request_runtime_ctx.get("prompt.abstract")
        if headline_data:
            request_messages.append(
                {"role": "user", "content": f"[主题及摘要]{to_instruction(headline_data)}"})
            request_messages.append({"role": "assistant", "content": "OK"})
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get("prompt.output")
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception(
                "[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            request_messages.append({"role": "user", "content": to_instruction(prompt_input_data)})
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[输入]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[补充信息]"] = str(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[处理规则]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[输出要求]"] = {
                        "TYPE": "可被解析的JSON字符串",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[输出要求]"] = str(prompt_output_data)
            request_messages.append(
                {"role": "user", "content": to_prompt_structure(prompt_dict, end="[输出]:\n")})
        return request_messages

    def generate_request_data(self):
        options = self.model_settings.get_trace_back("options", {})
        options = options if isinstance(options, dict) else {}
        access_token = self.model_settings.get_trace_back("auth", {})
        access_token = access_token if isinstance(access_token, dict) else {}
        request_data = {}
        # request type: chat
        if self.request_type == "chat":
            if "model" not in options:
                options["model"] = "ernie-4.0"
            if "aistudio" not in access_token and "qianfan" not in access_token:
                raise Exception(
                    f"[Request] ERNIE require 'access-token-for-aistudio or access-token-for-qianfan' when request type is '{self.request_type}'. Use .set_model_auth({{ 'aistudio': <YOUR-ACCESS-TOKEN-FOR-AISTUDIO> }} or {{ 'qianfan': <YOUR-ACCESS-TOKEN-FOR-QIANFAN> }}) to set.")
            api_type = next(iter(access_token))
            erniebot.api_type = api_type
            erniebot.access_token = access_token[api_type]
            messages = self.construct_request_messages()
            request_messages = []
            #system_prompt = ""
            for message in messages:
                if message["role"] == "system":
                    #system_prompt += f"{ message['content'] }\n"
                    message["role"] = "user"
                    request_messages.append(message)
                else:
                    request_messages.append(message)
            request_data = {
                "messages": format_request_messages(request_messages, self.model_settings),
                "stream": True,
                **options,
            }
            #if system_prompt != "" and self.request.settings.get_trace_back("retry_count", 0) > 0:
            #    request_data.update({ "system": system_prompt })
        # request type: embedding
        elif self.request_type == "embedding":
            if "model" not in options:
                options["model"] = "ernie-text-embedding"
            if "aistudio" not in access_token:
                raise Exception(
                    f"[Request] ERNIE require 'access-token-for-aistudio' when request type is '{self.request_type}'. Use .set_model_auth({{ 'aistudio': <YOUR-ACCESS-TOKEN-FOR-AISTUDIO> }}) to set.\n How to get your own access token? visit: https://github.com/PaddlePaddle/ERNIE-Bot-SDK/blob/develop/docs/authentication.md")
            erniebot.api_type = "aistudio"
            erniebot.access_token = access_token["aistudio"]
            content_input = self.request.request_runtime_ctx.get("prompt.input")
            if not isinstance(content_input, list):
                content_input = [content_input]
            request_data = {
                "input": content_input,
                **options,
            }
        # request type: image
        elif self.request_type == "image":
            if "model" not in options:
                options["model"] = "ernie-vilg-v2"
            if "yinian" not in access_token:
                raise Exception(
                    f"[Request] ERNIE require 'access-token-for-yinian' when request type is '{self.request_type}'. Use .set_model_auth({{ 'yinian': <YOUR-ACCESS-TOKEN-FOR-YINIAN> }}) to set.\n⚠️ Yinian Access Token is different from AIStudio Access Token!\n How to get your own access token? visit: https://github.com/PaddlePaddle/ERNIE-Bot-SDK/blob/develop/docs/authentication.md")
            erniebot.api_type = "yinian"
            erniebot.access_token = access_token["yinian"]
            prompt = self.request.request_runtime_ctx.get("prompt.input")
            output_requirement = self.request.request_runtime_ctx.get("prompt.output", {})
            if not isinstance(output_requirement, dict):
                output_requirement = {}
            if "width" not in output_requirement:
                output_requirement.update({"width": 512})
            if "height" not in output_requirement:
                output_requirement.update({"height": 512})
            request_data = {
                "prompt": prompt,
                **output_requirement,
                **options,
            }
        return request_data

    async def request_model(self, request_data: dict):
        client = self._create_client()
        response = await client.acreate(**request_data)
        return response

    async def broadcast_response(self, response_generator):
        if self.request_type == "chat":
            response_message = {"role": "assistant", "content": ""}
            full_response_message = {}
            async for part in response_generator:
                full_response_message = dict(part)
                delta = part["result"]
                response_message["content"] += delta
                yield ({"event": "response:delta_origin", "data": part})
                yield ({"event": "response:delta", "data": delta})
            full_response_message["result"] = response_message["content"]
            yield ({"event": "response:done_origin", "data": full_response_message})
            yield ({"event": "response:done", "data": response_message["content"]})
        elif self.request_type == "embedding":
            response = dict(response_generator)
            if response["rcode"] == 200:
                yield ({"event": "response:done_origin", "data": response})
                yield ({"event": "response:done", "data": response["rbody"]["data"]})
            else:
                raise Exception(f"[Request] ERNIE Error: {response}")
        elif self.request_type == "image":
            response = dict(response_generator)
            if response["data"]["task_status"] == "SUCCESS":
                yield ({"event": "response:done_origin", "data": response["data"]})
                yield ({"event": "response:done",
                        "data": response["data"]["sub_task_result_list"][0]["final_image_list"][0][
                            "img_url"]})
            else:
                raise Exception(f"[Request] ERNIE Error: {response['data']}")

    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }


def export():
    return ("ERNIE", Ernie)



from typing import TYPE_CHECKING
from .Repository import CheckpointRepository
from .Snapshot import Snapshot
from ..lib.constants import DEFAULT_CHECKPOINT_NAME
from ..utils.runner import run_async

if TYPE_CHECKING:
    from .State import RuntimeState

class Checkpoint:
  """ Checkpoint 管理逻辑，包含快照数据存储、回溯、恢复等 """

  def __init__(self, checkpoint_id, repository=CheckpointRepository(), default_snapshot_name: str=None) -> None:
    # 默认存储的快照名
    self.default_snapshot_name = default_snapshot_name or DEFAULT_CHECKPOINT_NAME
    self.checkpoint_id = checkpoint_id
    self.repository: CheckpointRepository = repository
    self.active_snapshot: Snapshot = None
    self._disabled = False
  
  def use(self, name: str):
    """使用指定的快照名（存储、回滚、删除默认都会使用该快照名）"""
    self.default_snapshot_name = name
    return self
  
  def disable(self):
    self._disabled = True
    return self

  def save(self, state: 'RuntimeState', name=None, time=None, **args):
    """将某个状态存储到快照记录中"""
    if self._disabled:
      return self
    return run_async(self.save_async(state=state, name=name, time=time,  **args))

  async def save_async(self, state: 'RuntimeState', name=None, time=None):
    """将某个状态存储到快照记录中"""
    if self._disabled:
      return self
    snapshot_name = name or self.default_snapshot_name
    snapshot = Snapshot(name=snapshot_name, state=state, time=time)
    await self.repository.save(checkpoint_id=self.checkpoint_id, name=snapshot_name, data=snapshot.export())
    return self
  
  def remove(self, name=None):
    if self._disabled:
      return self
    snapshot_name = name or self.default_snapshot_name
    return run_async(self.remove_async(name=snapshot_name))

  async def remove_async(self, name=None):
    """删除指定的快照存储"""
    if self._disabled:
      return self
    snapshot_name = name or self.default_snapshot_name
    await self.repository.remove(checkpoint_id=self.checkpoint_id, name=snapshot_name)
    return self

  def rollback(self, name=None, silence=False, **args):
    """回滚到指定checkpoint"""
    if self._disabled:
      return self
    snapshot_name = name or self.default_snapshot_name
    run_async(self.rollback_async(name=snapshot_name, silence=silence, **args))

  async def rollback_async(self, name=None, silence=False):
    """回滚到指定checkpoint"""
    if self._disabled:
      return self
    snapshot_name = name or self.default_snapshot_name
    snapshot_raw_data = await self.repository.get(checkpoint_id=self.checkpoint_id, name=snapshot_name)
    if not snapshot_raw_data:
      # 静默态，无需抛错
      if silence:
        return self
      raise ValueError(
          f'Attempt to roll back to a non-existent checkpoint record named "{name}".')

    self.active_snapshot = Snapshot(**snapshot_raw_data)
    return self
  
  def get_active_snapshot(self):
    """返回当前激活的快照"""
    return self.active_snapshot



"""
Tokens for lexer
"""

# Token constants
TOKEN_EOF = 0  # end-of-file
TOKEN_IGNORED = 1  # \t', '\n', '\v', '\f', '\r', ' '
TOKEN_LEFT_BRACKET = 2  # [
TOKEN_RIGHT_BRACKET = 3  # ]
TOKEN_LEFT_BRACE = 4  # {
TOKEN_RIGHT_BRACE = 5  # }
TOKEN_COLON = 6  # :
TOKEN_DOT = 7  # .
TOKEN_COMMA = 8  # ,
TOKEN_QUOTE = 9  # "
TOKEN_ESCAPE_CHARACTER = 10  # \
TOKEN_SLASH = 11  # /
TOKEN_NEGATIVE = 12  # -
TOKEN_NULL = 13  # null
TOKEN_TRUE = 14  # true
TOKEN_FALSE = 15  # false
TOKEN_ALPHABET_LOWERCASE_A = 16  # a
TOKEN_ALPHABET_LOWERCASE_B = 17  # b
TOKEN_ALPHABET_LOWERCASE_C = 18  # c
TOKEN_ALPHABET_LOWERCASE_D = 19  # d
TOKEN_ALPHABET_LOWERCASE_E = 20  # e
TOKEN_ALPHABET_LOWERCASE_F = 21  # f
TOKEN_ALPHABET_LOWERCASE_L = 22  # l
TOKEN_ALPHABET_LOWERCASE_N = 23  # n
TOKEN_ALPHABET_LOWERCASE_R = 24  # r
TOKEN_ALPHABET_LOWERCASE_S = 25  # s
TOKEN_ALPHABET_LOWERCASE_T = 26  # t
TOKEN_ALPHABET_LOWERCASE_U = 27  # u
TOKEN_ALPHABET_UPPERCASE_A = 28  # A
TOKEN_ALPHABET_UPPERCASE_B = 29  # B
TOKEN_ALPHABET_UPPERCASE_C = 30  # C
TOKEN_ALPHABET_UPPERCASE_D = 31  # D
TOKEN_ALPHABET_UPPERCASE_E = 32  # E
TOKEN_ALPHABET_UPPERCASE_F = 33  # F
TOKEN_NUMBER = 34  # number
TOKEN_NUMBER_0 = 35  # 0
TOKEN_NUMBER_1 = 36  # 1
TOKEN_NUMBER_2 = 37  # 2
TOKEN_NUMBER_3 = 38  # 3
TOKEN_NUMBER_4 = 39  # 4
TOKEN_NUMBER_5 = 40  # 5
TOKEN_NUMBER_6 = 41  # 6
TOKEN_NUMBER_7 = 42  # 7
TOKEN_NUMBER_8 = 43  # 8
TOKEN_NUMBER_9 = 44  # 9
TOKEN_OTHERS = 45  # anything else in json

# Token Symbols
TOKEN_LEFT_BRACKET_SYMBOL = "["
TOKEN_RIGHT_BRACKET_SYMBOL = "]"
TOKEN_LEFT_BRACE_SYMBOL = "{"
TOKEN_RIGHT_BRACE_SYMBOL = "}"
TOKEN_COLON_SYMBOL = ":"
TOKEN_DOT_SYMBOL = "."
TOKEN_COMMA_SYMBOL = ","
TOKEN_QUOTE_SYMBOL = '"'
TOKEN_ESCAPE_CHARACTER_SYMBOL = "\\"
TOKEN_SLASH_SYMBOL = "/"
TOKEN_NEGATIVE_SYMBOL = "-"
TOKEN_ALPHABET_LOWERCASE_A_SYMBOL = "a"
TOKEN_ALPHABET_LOWERCASE_B_SYMBOL = "b"
TOKEN_ALPHABET_LOWERCASE_C_SYMBOL = "c"
TOKEN_ALPHABET_LOWERCASE_D_SYMBOL = "d"
TOKEN_ALPHABET_LOWERCASE_E_SYMBOL = "e"
TOKEN_ALPHABET_LOWERCASE_F_SYMBOL = "f"
TOKEN_ALPHABET_LOWERCASE_L_SYMBOL = "l"
TOKEN_ALPHABET_LOWERCASE_N_SYMBOL = "n"
TOKEN_ALPHABET_LOWERCASE_R_SYMBOL = "r"
TOKEN_ALPHABET_LOWERCASE_S_SYMBOL = "s"
TOKEN_ALPHABET_LOWERCASE_T_SYMBOL = "t"
TOKEN_ALPHABET_LOWERCASE_U_SYMBOL = "u"
TOKEN_ALPHABET_UPPERCASE_A_SYMBOL = "A"
TOKEN_ALPHABET_UPPERCASE_B_SYMBOL = "B"
TOKEN_ALPHABET_UPPERCASE_C_SYMBOL = "C"
TOKEN_ALPHABET_UPPERCASE_D_SYMBOL = "D"
TOKEN_ALPHABET_UPPERCASE_E_SYMBOL = "E"
TOKEN_ALPHABET_UPPERCASE_F_SYMBOL = "F"
TOKEN_NUMBER_0_SYMBOL = "0"
TOKEN_NUMBER_1_SYMBOL = "1"
TOKEN_NUMBER_2_SYMBOL = "2"
TOKEN_NUMBER_3_SYMBOL = "3"
TOKEN_NUMBER_4_SYMBOL = "4"
TOKEN_NUMBER_5_SYMBOL = "5"
TOKEN_NUMBER_6_SYMBOL = "6"
TOKEN_NUMBER_7_SYMBOL = "7"
TOKEN_NUMBER_8_SYMBOL = "8"
TOKEN_NUMBER_9_SYMBOL = "9"


# Token symbol map
token_symbol_map = {
    TOKEN_EOF: "EOF",
    TOKEN_LEFT_BRACKET: "[",
    TOKEN_RIGHT_BRACKET: "]",
    TOKEN_LEFT_BRACE: "{",
    TOKEN_RIGHT_BRACE: "}",
    TOKEN_COLON: ":",
    TOKEN_DOT: ".",
    TOKEN_COMMA: ",",
    TOKEN_QUOTE: '"',
    TOKEN_ESCAPE_CHARACTER: "\\",
    TOKEN_SLASH: "/",
    TOKEN_NEGATIVE: "-",
    TOKEN_NULL: "null",
    TOKEN_TRUE: "true",
    TOKEN_FALSE: "false",
    TOKEN_ALPHABET_LOWERCASE_A: "a",
    TOKEN_ALPHABET_LOWERCASE_B: "b",
    TOKEN_ALPHABET_LOWERCASE_C: "c",
    TOKEN_ALPHABET_LOWERCASE_D: "d",
    TOKEN_ALPHABET_LOWERCASE_E: "e",
    TOKEN_ALPHABET_LOWERCASE_F: "f",
    TOKEN_ALPHABET_LOWERCASE_L: "l",
    TOKEN_ALPHABET_LOWERCASE_N: "n",
    TOKEN_ALPHABET_LOWERCASE_R: "r",
    TOKEN_ALPHABET_LOWERCASE_S: "s",
    TOKEN_ALPHABET_LOWERCASE_T: "t",
    TOKEN_ALPHABET_LOWERCASE_U: "u",
    TOKEN_ALPHABET_UPPERCASE_A: "A",
    TOKEN_ALPHABET_UPPERCASE_B: "B",
    TOKEN_ALPHABET_UPPERCASE_C: "C",
    TOKEN_ALPHABET_UPPERCASE_D: "D",
    TOKEN_ALPHABET_UPPERCASE_E: "E",
    TOKEN_ALPHABET_UPPERCASE_F: "F",
    TOKEN_NUMBER_0: "0",
    TOKEN_NUMBER_1: "1",
    TOKEN_NUMBER_2: "2",
    TOKEN_NUMBER_3: "3",
    TOKEN_NUMBER_4: "4",
    TOKEN_NUMBER_5: "5",
    TOKEN_NUMBER_6: "6",
    TOKEN_NUMBER_7: "7",
    TOKEN_NUMBER_8: "8",
    TOKEN_NUMBER_9: "9",
}



import uuid

class IdGenerator(object):
    def __init__(self, namespace: str):
        self.namespace = namespace
        self.namespace_uuid = uuid.uuid5(uuid.NAMESPACE_DNS, self.namespace)

    def create(self):
        individual_id = uuid.uuid4()
        return str(uuid.uuid5(self.namespace_uuid, str(individual_id)))


import os
import importlib

def scan():
    type_list = []
    create_dict = {}
    dir_path = os.path.dirname(os.path.abspath(__file__))
    for item in os.listdir(dir_path):
        if item.endswith(".py") and item != "__init__.py":
            chunk_type_name = item[:-3]
            type_list.append(chunk_type_name)
            preset_chunk = importlib.import_module(f".{ chunk_type_name }", package = __package__)
            if not hasattr(preset_chunk, "create"):
                raise Exception(f"Preset chunk '{ chunk_type_name }' must state a `create()` function.")
            create_dict.update({ chunk_type_name: getattr(preset_chunk, "create") })
    return {
        "type_list": type_list,
        "create": create_dict,
    }

preset_chunks = scan()


"""
streaming-json-py main lexer method
This method will 
"""

from Agently.utils.lexer import lexer_tokens, lexer_helper

class Lexer:
    """
    lexer for json fragment
    """

    def __init__(self):
        self.json_content = []  # input JSON content
        self.padding_content = (
            []
        )  # padding content for ignored characters and escape characters, etc.
        self.json_segment = ""  # appended JSON segment by the AppendString() method.
        self.token_stack = []  # token stack for input JSON
        self.mirror_token_stack = []  # token stack for auto-completed tokens

    def __get_top_token_on_stack(self):
        """
        get token on the stack top
        """
        if not self.token_stack:
            return lexer_tokens.TOKEN_EOF
        return self.token_stack[-1]

    def __get_top_token_on_mirror_stack(self):
        """
        get token on the mirror stack top
        """
        if not self.mirror_token_stack:
            return lexer_tokens.TOKEN_EOF
        return self.mirror_token_stack[-1]

    def __pop_token_stack(self):
        """
        pop token on the stack top
        """
        if not self.token_stack:
            return lexer_tokens.TOKEN_EOF
        return self.token_stack.pop()

    def __pop_mirror_token_stack(self):
        """
        pop token on the mirror stack top
        """
        if not self.mirror_token_stack:
            return lexer_tokens.TOKEN_EOF
        return self.mirror_token_stack.pop()

    def __push_token_stack(self, token):
        """
        push token into the stack
        """
        self.token_stack.append(token)

    def __push_mirror_token_stack(self, token):
        """
        push token into the mirror stack
        """
        self.mirror_token_stack.append(token)

    def __dump_mirror_token_stack_to_string(self):
        """
        convert mirror stack token into string
        """
        return "".join(
            [
                lexer_tokens.token_symbol_map[x]
                for x in reversed(self.mirror_token_stack)
            ]
        )

    def __skip_json_segment(self, n):
        """
        skip JSON segment by length n
        """
        self.json_segment = self.json_segment[n:]

    def __push_negative_into_json_content(self):
        """
        push negative symbol `-` into JSON content
        """
        self.json_content.append(lexer_tokens.TOKEN_NEGATIVE_SYMBOL)

    def __push_byte_into_padding_content(self, b):
        """
        push byte into JSON content by given
        """
        self.padding_content.append(b)

    def __append_padding_content_to_json_content(self):
        """
        append padding content into JSON content
        """
        self.json_content.extend(self.padding_content)
        self.padding_content = []

    def __have_padding_content(self):
        """
        check if padding content is empty
        """
        return bool(self.padding_content)

    def __clean_padding_content(self):
        """
        set padding content to empty
        """
        self.padding_content = []

    def __stream_stopped_in_an_object_key_start(self) -> bool:
        """
        check if JSON stream stopped at an object properity's key start, like `{"`
        """
        # `{`, `"` in stack, or `,`, `"` in stack
        case1 = [lexer_tokens.TOKEN_LEFT_BRACE, lexer_tokens.TOKEN_QUOTE]
        case2 = [lexer_tokens.TOKEN_COMMA, lexer_tokens.TOKEN_QUOTE]
        #  `}` in mirror stack
        case3 = [lexer_tokens.TOKEN_RIGHT_BRACE]
        return (
            lexer_helper.match_stack(self.token_stack, case1)
            or lexer_helper.match_stack(self.token_stack, case2)
        ) and lexer_helper.match_stack(self.mirror_token_stack, case3)

    def __stream_stopped_in_an_object_key_end(self) -> bool:
        """
        check if JSON stream stopped in an object properity's key, like `{"field`
        """
        # // `{`, `"`, `"` in stack, or `,`, `"`, `"` in stack
        case1 = [
            lexer_tokens.TOKEN_LEFT_BRACE,
            lexer_tokens.TOKEN_QUOTE,
            lexer_tokens.TOKEN_QUOTE,
        ]
        case2 = [
            lexer_tokens.TOKEN_COMMA,
            lexer_tokens.TOKEN_QUOTE,
            lexer_tokens.TOKEN_QUOTE,
        ]
        # // `"`, `:`, `n`, `u`, `l`, `l`, `}` in mirror stack
        case3 = [
            lexer_tokens.TOKEN_RIGHT_BRACE,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N,
            lexer_tokens.TOKEN_COLON,
            lexer_tokens.TOKEN_QUOTE,
        ]
        return (
            lexer_helper.match_stack(self.token_stack, case1)
            or lexer_helper.match_stack(self.token_stack, case2)
        ) and lexer_helper.match_stack(self.mirror_token_stack, case3)

    def __stream_stopped_in_an_object_string_value_start(self) -> bool:
        """
        check if JSON stream stopped in an object properity's value start,
        like `{"field": "`
        """

        # `:`, `"` in stack
        case1 = [lexer_tokens.TOKEN_COLON, lexer_tokens.TOKEN_QUOTE]
        # // `n`, `u`, `l`, `l`, `}` in mirror stack
        case2 = [
            lexer_tokens.TOKEN_RIGHT_BRACE,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N,
        ]
        return lexer_helper.match_stack(
            self.token_stack, case1
        ) and lexer_helper.match_stack(self.mirror_token_stack, case2)

    def __stream_stopped_in_an_object_value_end(self) -> bool:
        """
        check if JSON stream stopped in an object properity's value finish,
        like `{"field": "value"`
        """
        # `"`, `}` left
        tokens = [lexer_tokens.TOKEN_RIGHT_BRACE, lexer_tokens.TOKEN_QUOTE]
        return lexer_helper.match_stack(self.mirror_token_stack, tokens)

    def __stream_stopped_in_an_object_array_value_start(self) -> bool:
        """
        check if JSON stream stopped in an object properity's value start by array,
        like `{"field":[`
        """
        # `:`, `[` in stack
        case1 = [lexer_tokens.TOKEN_COLON, lexer_tokens.TOKEN_LEFT_BRACKET]
        # `n`, `u`, `l`, `l`, `}` in mirror stack
        case2 = [
            lexer_tokens.TOKEN_RIGHT_BRACE,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N,
        ]
        return lexer_helper.match_stack(
            self.token_stack, case1
        ) and lexer_helper.match_stack(self.mirror_token_stack, case2)

    def __stream_stopped_in_an_object_object_value_start(self) -> bool:
        """
        check if JSON stream stopped in an object properity's value start by array,
        like `{"field":{`
        """
        # `:`, `{` in stack
        case1 = [lexer_tokens.TOKEN_COLON, lexer_tokens.TOKEN_LEFT_BRACE]
        # `n`, `u`, `l`, `l`, `}` in mirror stack
        case2 = [
            lexer_tokens.TOKEN_RIGHT_BRACE,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N,
        ]
        return lexer_helper.match_stack(
            self.token_stack, case1
        ) and lexer_helper.match_stack(self.mirror_token_stack, case2)

    def __stream_stopped_in_an_object_negative_number_value_start(self) -> bool:
        """
        check if JSON stream stopped in an object properity's negative number value,
        like `:-`
        """
        # `:`, `-` in stack
        case1 = [lexer_tokens.TOKEN_COLON, lexer_tokens.TOKEN_NEGATIVE]
        return lexer_helper.match_stack(self.token_stack, case1)

    def __stream_stopped_in_a_negative_number_value_start(self) -> bool:
        """
        check if JSON stream stopped in an object properity's negative number value,
        like `-`
        """
        # `-` in stack
        case1 = [lexer_tokens.TOKEN_NEGATIVE]
        # `0`in mirror stack
        case2 = [lexer_tokens.TOKEN_NUMBER_0]
        return lexer_helper.match_stack(
            self.token_stack, case1
        ) and lexer_helper.match_stack(self.mirror_token_stack, case2)

    def __stream_stopped_in_an_array(self) -> bool:
        """
        check if JSON stream stopped in an array
        """
        return (
            self.__get_top_token_on_mirror_stack() == lexer_tokens.TOKEN_RIGHT_BRACKET
        )

    def __stream_stopped_in_an_array_string_value_end(self) -> bool:
        """
        check if JSON stream stopped in an array's string value end, like `["value"`
        """
        # `"`, `"` in stack
        case1 = [lexer_tokens.TOKEN_QUOTE, lexer_tokens.TOKEN_QUOTE]
        # `"`, `]` in mirror stack
        case2 = [lexer_tokens.TOKEN_RIGHT_BRACKET, lexer_tokens.TOKEN_QUOTE]
        return lexer_helper.match_stack(
            self.token_stack, case1
        ) and lexer_helper.match_stack(self.mirror_token_stack, case2)

    def __stream_stopped_in_an_object_null_value_placeholder_start(self) -> bool:
        """
        check if JSON stream stopped in an object properity's value start by array,
        like `{"field":{`
        """
        # `n`, `u`, `l`, `l`, `}` in mirror stack
        case1 = [
            lexer_tokens.TOKEN_RIGHT_BRACE,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N,
        ]
        return lexer_helper.match_stack(self.mirror_token_stack, case1)

    def __stream_stopped_in_a_string(self) -> bool:
        """
        check if JSON stream stopped in a string, like `""`
        """
        return (
            self.__get_top_token_on_stack() == lexer_tokens.TOKEN_QUOTE
            and self.__get_top_token_on_mirror_stack() == lexer_tokens.TOKEN_QUOTE
        )

    def __stream_stopped_in_an_string_unicode_escape(self) -> bool:
        """
        check if JSON stream stopped in a string's unicode escape, like `\u0001`
        """
        # `\`, `u` in stack
        case1 = [
            lexer_tokens.TOKEN_ESCAPE_CHARACTER,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
        ]
        # `"` in mirror stack
        case2 = [lexer_tokens.TOKEN_QUOTE]
        return lexer_helper.match_stack(
            self.token_stack, case1
        ) and lexer_helper.match_stack(self.mirror_token_stack, case2)

    def __stream_stopped_in_a_number(self) -> bool:
        """
        check if JSON stream stopped in a number, like `[0-9]`
        """
        return self.__get_top_token_on_stack() == lexer_tokens.TOKEN_NUMBER

    def __stream_stopped_in_a_number_decimal_part(self) -> bool:
        """
        check if JSON stream stopped in a number first decimal place, like `.?`
        """
        # `.`, lexer_tokens.TOKEN_NUMBER in stack
        return self.__get_top_token_on_stack() == lexer_tokens.TOKEN_DOT

    def __stream_stopped_in_a_number_decimal_part_middle(self) -> bool:
        """
        check if JSON stream stopped in a number other decimal place (except first place),
        like `.[0-9]?`
        """
        case1 = [lexer_tokens.TOKEN_DOT, lexer_tokens.TOKEN_NUMBER]
        return lexer_helper.match_stack(self.token_stack, case1)

    def __stream_stopped_with_leading_escape_character(self) -> bool:
        """
        check if JSON stream stopped in escape character, like \
        """
        return self.__get_top_token_on_stack() == lexer_tokens.TOKEN_ESCAPE_CHARACTER

    def __match_token(self):
        """
        lexer match JSON token method, convert JSON segment to JSON tokens
        """
        # Segment end
        if len(self.json_segment) == 0:
            return lexer_tokens.TOKEN_EOF, 0

        token_symbol = self.json_segment[0]

        # Check if ignored token
        if lexer_helper.is_ignore_token(token_symbol):
            self.__skip_json_segment(1)
            return lexer_tokens.TOKEN_IGNORED, token_symbol
        # Match token
        token_mapping = {
            lexer_tokens.TOKEN_LEFT_BRACKET_SYMBOL: lexer_tokens.TOKEN_LEFT_BRACKET,
            lexer_tokens.TOKEN_RIGHT_BRACKET_SYMBOL: lexer_tokens.TOKEN_RIGHT_BRACKET,
            lexer_tokens.TOKEN_LEFT_BRACE_SYMBOL: lexer_tokens.TOKEN_LEFT_BRACE,
            lexer_tokens.TOKEN_RIGHT_BRACE_SYMBOL: lexer_tokens.TOKEN_RIGHT_BRACE,
            lexer_tokens.TOKEN_COLON_SYMBOL: lexer_tokens.TOKEN_COLON,
            lexer_tokens.TOKEN_DOT_SYMBOL: lexer_tokens.TOKEN_DOT,
            lexer_tokens.TOKEN_COMMA_SYMBOL: lexer_tokens.TOKEN_COMMA,
            lexer_tokens.TOKEN_QUOTE_SYMBOL: lexer_tokens.TOKEN_QUOTE,
            lexer_tokens.TOKEN_ESCAPE_CHARACTER_SYMBOL: lexer_tokens.TOKEN_ESCAPE_CHARACTER,
            lexer_tokens.TOKEN_SLASH_SYMBOL: lexer_tokens.TOKEN_SLASH,
            lexer_tokens.TOKEN_NEGATIVE_SYMBOL: lexer_tokens.TOKEN_NEGATIVE,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_A_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_A,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_B_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_B,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_C_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_C,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_D_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_D,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_F_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_F,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_R_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_R,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_S_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_S,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_T_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_T,
            lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U_SYMBOL: lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
            lexer_tokens.TOKEN_ALPHABET_UPPERCASE_A_SYMBOL: lexer_tokens.TOKEN_ALPHABET_UPPERCASE_A,
            lexer_tokens.TOKEN_ALPHABET_UPPERCASE_B_SYMBOL: lexer_tokens.TOKEN_ALPHABET_UPPERCASE_B,
            lexer_tokens.TOKEN_ALPHABET_UPPERCASE_C_SYMBOL: lexer_tokens.TOKEN_ALPHABET_UPPERCASE_C,
            lexer_tokens.TOKEN_ALPHABET_UPPERCASE_D_SYMBOL: lexer_tokens.TOKEN_ALPHABET_UPPERCASE_D,
            lexer_tokens.TOKEN_ALPHABET_UPPERCASE_E_SYMBOL: lexer_tokens.TOKEN_ALPHABET_UPPERCASE_E,
            lexer_tokens.TOKEN_ALPHABET_UPPERCASE_F_SYMBOL: lexer_tokens.TOKEN_ALPHABET_UPPERCASE_F,
            lexer_tokens.TOKEN_NUMBER_0_SYMBOL: lexer_tokens.TOKEN_NUMBER_0,
            lexer_tokens.TOKEN_NUMBER_1_SYMBOL: lexer_tokens.TOKEN_NUMBER_1,
            lexer_tokens.TOKEN_NUMBER_2_SYMBOL: lexer_tokens.TOKEN_NUMBER_2,
            lexer_tokens.TOKEN_NUMBER_3_SYMBOL: lexer_tokens.TOKEN_NUMBER_3,
            lexer_tokens.TOKEN_NUMBER_4_SYMBOL: lexer_tokens.TOKEN_NUMBER_4,
            lexer_tokens.TOKEN_NUMBER_5_SYMBOL: lexer_tokens.TOKEN_NUMBER_5,
            lexer_tokens.TOKEN_NUMBER_6_SYMBOL: lexer_tokens.TOKEN_NUMBER_6,
            lexer_tokens.TOKEN_NUMBER_7_SYMBOL: lexer_tokens.TOKEN_NUMBER_7,
            lexer_tokens.TOKEN_NUMBER_8_SYMBOL: lexer_tokens.TOKEN_NUMBER_8,
            lexer_tokens.TOKEN_NUMBER_9_SYMBOL: lexer_tokens.TOKEN_NUMBER_9,
        }

        token_result = token_mapping.get(token_symbol, lexer_tokens.TOKEN_OTHERS)
        self.__skip_json_segment(1)
        return token_result, token_symbol

    def append_string(
        self,
        string: str,
    ):
        """
        append JSON string to current JSON stream content
        this method will traversal all token and generate mirror token for complete full JSON
        """

        self.json_segment = string
        while True:
            token, token_symbol = self.__match_token()

            if token == lexer_tokens.TOKEN_EOF:
                # nothing to do with TOKEN_EOF
                pass
            elif token == lexer_tokens.TOKEN_IGNORED:
                if self.__stream_stopped_in_a_string():
                    self.json_content += token_symbol
                    continue
                self.__push_byte_into_padding_content(token_symbol)
            elif token == lexer_tokens.TOKEN_OTHERS:
                # check if json stream stopped with padding content
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                # write current token symbol to JSON content
                self.json_content += token_symbol
            elif token == lexer_tokens.TOKEN_LEFT_BRACKET:
                # check if json stream stopped with padding content
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                self.json_content += token_symbol
                if self.__stream_stopped_in_a_string():
                    continue
                self.__push_token_stack(token)
                if self.__stream_stopped_in_an_object_array_value_start():
                    # pop `n`, `u`, `l`, `l` from mirror stack
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                # push `]` into mirror stack
                self.__push_mirror_token_stack(lexer_tokens.TOKEN_RIGHT_BRACKET)
            elif token == lexer_tokens.TOKEN_RIGHT_BRACKET:
                if self.__stream_stopped_in_a_string():
                    self.json_content += token_symbol
                    continue
                # check if json stream stopped with padding content
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                # write current token symbol to JSON content
                self.json_content += token_symbol
                # push `]` into stack
                self.__push_token_stack(token)
                # pop `]` from mirror stack
                self.__pop_mirror_token_stack()
            elif token == lexer_tokens.TOKEN_LEFT_BRACE:
                # check if json stream stopped with padding content
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                # write current token symbol to JSON content
                self.json_content += token_symbol
                if self.__stream_stopped_in_a_string():
                    continue
                self.__push_token_stack(token)
                if self.__stream_stopped_in_an_object_object_value_start():
                    # pop `n`, `u`, `l`, `l` from mirror stack
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                # push `}` into mirror stack
                self.__push_mirror_token_stack(lexer_tokens.TOKEN_RIGHT_BRACE)
            elif token == lexer_tokens.TOKEN_RIGHT_BRACE:
                if self.__stream_stopped_in_a_string():
                    self.json_content += token_symbol
                    continue
                # check if json stream stopped with padding content
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                self.json_content += token_symbol
                # push `}` into stack
                self.__push_token_stack(token)
                # pop `}` from mirror stack
                self.__pop_mirror_token_stack()
            elif token == lexer_tokens.TOKEN_QUOTE:
                # check if escape quote `\"`
                if self.__stream_stopped_with_leading_escape_character():
                    # push padding escape character `\` into JSON content
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                    # write current token symbol to JSON content
                    self.json_content += token_symbol
                    # pop `\` from  stack
                    self.__pop_token_stack()
                    continue
                # check if json stream stopped with padding content
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()

                # write current token symbol to JSON content
                self.json_content += token_symbol
                self.__push_token_stack(token)
                if self.__stream_stopped_in_an_array():
                    # push `"` into mirror stack
                    self.__push_mirror_token_stack(lexer_tokens.TOKEN_QUOTE)
                elif self.__stream_stopped_in_an_array_string_value_end():
                    # pop `"` from mirror stack
                    self.__pop_mirror_token_stack()
                elif self.__stream_stopped_in_an_object_key_start():
                    #  check if stopped in key of object's properity or value of object's properity
                    #  push `"`, `:`, `n`, `u`, `l`, `l` into mirror stack
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N
                    )
                    self.__push_mirror_token_stack(lexer_tokens.TOKEN_COLON)
                    self.__push_mirror_token_stack(lexer_tokens.TOKEN_QUOTE)
                elif self.__stream_stopped_in_an_object_key_end():
                    # check if stopped in key of object's properity or value of object's properity
                    # pop `"` from mirror stack
                    self.__pop_mirror_token_stack()
                elif self.__stream_stopped_in_an_object_string_value_start():
                    # pop `n`, `u`, `l`, `l` from mirror stack
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    # push `"` into mirror stack
                    self.__push_mirror_token_stack(lexer_tokens.TOKEN_QUOTE)
                elif self.__stream_stopped_in_an_object_value_end():
                    # pop `"` from mirror stack
                    self.__pop_mirror_token_stack()
                else:
                    return "Invalid quote token in JSON stream"
            elif token == lexer_tokens.TOKEN_COLON:
                if self.__stream_stopped_in_a_string():
                    self.json_content += token_symbol
                    continue
                # check if json stream stopped with padding content
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                # write current token symbol to JSON content
                self.json_content += token_symbol
                self.__push_token_stack(token)
                # pop `:` from mirror stack
                self.__pop_mirror_token_stack()
            elif token == lexer_tokens.TOKEN_ALPHABET_LOWERCASE_A:
                # as hex in unicode
                if self.__stream_stopped_in_an_string_unicode_escape():
                    self.__push_byte_into_padding_content(token_symbol)
                    # check if unicode escape is full length
                    if len(self.padding_content) == 6:
                        self.__append_padding_content_to_json_content()
                        self.__clean_padding_content()
                        # pop `\`, `u` from stack
                        self.__pop_token_stack()
                        self.__pop_token_stack()
                    continue
                # write current token symbol to JSON content
                self.json_content += token_symbol
                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

                # check if `f` in token stack and `a`, `l`, `s`, `e in mirror stack
                def it_is_part_of_token_false():
                    left = [lexer_tokens.TOKEN_ALPHABET_LOWERCASE_F]
                    right = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_S,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_A,
                    ]
                    return lexer_helper.match_stack(
                        self.token_stack, left
                    ) and lexer_helper.match_stack(self.mirror_token_stack, right)

                if not it_is_part_of_token_false():
                    continue

                self.__push_token_stack(token)
                self.__pop_mirror_token_stack()
            elif token == lexer_tokens.TOKEN_ALPHABET_LOWERCASE_B:
                # as hex in unicode
                if self.__stream_stopped_in_an_string_unicode_escape():
                    self.__push_byte_into_padding_content(token_symbol)
                    # check if unicode escape is full length
                    if len(self.padding_content) == 6:
                        self.__append_padding_content_to_json_content()
                        self.__clean_padding_content()
                        # pop `\`, `u` from stack
                        self.__pop_token_stack()
                        self.__pop_token_stack()
                    continue

                # \b escape `\`, `b`
                if self.__stream_stopped_with_leading_escape_character():
                    # push padding escape character `\` into JSON content
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                    # write current token symbol to JSON content
                    self.json_content += token_symbol
                    # pop `\` from  stack
                    self.__pop_token_stack()
                    continue

                # write current token symbol to JSON content
                self.json_content += token_symbol

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

            elif token == lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E:
                # as hex in unicode
                if self.__stream_stopped_in_an_string_unicode_escape():
                    self.__push_byte_into_padding_content(token_symbol)
                    # check if unicode escape is full length
                    if len(self.padding_content) == 6:
                        self.__append_padding_content_to_json_content()
                        self.__clean_padding_content()
                        # pop `\`, `u` from stack
                        self.__pop_token_stack()
                        self.__pop_token_stack()
                    continue

                # check if in a number, as `e` (exponent) in scientific notation
                if self.__stream_stopped_in_a_number_decimal_part_middle():
                    self.__push_byte_into_padding_content(token_symbol)
                    continue

                # write current token symbol to JSON content
                self.json_content += token_symbol

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

                # check if `f`, `a`, `l`, `s` in token stack and `e` in mirror stack
                def it_is_part_of_token_false():
                    left = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_F,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_A,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_S,
                    ]
                    right = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E,
                    ]
                    return lexer_helper.match_stack(
                        self.token_stack, left
                    ) and lexer_helper.match_stack(self.mirror_token_stack, right)

                # check if `t`, `r`, `u` in token stack and `e` in mirror stack
                def it_is_part_of_token_true():
                    left = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_T,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_R,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
                    ]
                    right = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E,
                    ]
                    return lexer_helper.match_stack(
                        self.token_stack, left
                    ) and lexer_helper.match_stack(self.mirror_token_stack, right)

                if not it_is_part_of_token_false() and not it_is_part_of_token_true():
                    continue

                self.__push_token_stack(token)
                self.__pop_mirror_token_stack()
            elif token == lexer_tokens.TOKEN_ALPHABET_LOWERCASE_F:
                # as hex in unicode
                if self.__stream_stopped_in_an_string_unicode_escape():
                    self.__push_byte_into_padding_content(token_symbol)
                    # check if unicode escape is full length
                    if len(self.padding_content) == 6:
                        self.__append_padding_content_to_json_content()
                        self.__clean_padding_content()
                        # pop `\`, `u` from stack
                        self.__pop_token_stack()
                        self.__pop_token_stack()
                    continue

                # \f escape `\`, `f`
                if self.__stream_stopped_with_leading_escape_character():
                    # push padding escape character `\` into JSON content
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                    # write current token symbol to JSON content
                    self.json_content.append(token_symbol)
                    # pop `\` from  stack
                    self.__pop_token_stack()
                    continue

                # check if json stream stopped with padding content, like case `[true , f`
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()

                # write current token symbol to JSON content
                self.json_content.append(token_symbol)

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

                # push `f` into stack
                self.__push_token_stack(token)
                if self.__stream_stopped_in_an_array():
                    #  in array
                    #  push `a`, `l`, `s`, `e`
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_S
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_A
                    )
                else:
                    # in object
                    # pop `n`, `u`, `l`, `l`
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    # push `a`, `l`, `s`, `e`
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_S
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_A
                    )

            elif token == lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L:
                # write current token symbol to JSON content
                self.json_content.append(token_symbol)
                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

                # check if `f`, `a` in token stack and, `l`, `s`, `e` in mirror stack
                def it_is_part_of_token_false():
                    left = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_F,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_A,
                    ]
                    right = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_S,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
                    ]
                    return lexer_helper.match_stack(
                        self.token_stack, left
                    ) and lexer_helper.match_stack(self.mirror_token_stack, right)

                # check if `n`, `u` in token stack and `l`, `l` in mirror stack
                def it_is_part_of_token_null1():
                    left = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
                    ]
                    right = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
                    ]
                    return lexer_helper.match_stack(
                        self.token_stack, left
                    ) and lexer_helper.match_stack(self.mirror_token_stack, right)

                # check if `n`, `u`, `l` in token stack and `l` in mirror stack
                def it_is_part_of_token_null2():
                    left = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
                    ]
                    right = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
                    ]
                    return lexer_helper.match_stack(
                        self.token_stack, left
                    ) and lexer_helper.match_stack(self.mirror_token_stack, right)

                if (
                    not it_is_part_of_token_false()
                    and not it_is_part_of_token_null1()
                    and not it_is_part_of_token_null2()
                ):
                    continue

                self.__push_token_stack(token)
                self.__pop_mirror_token_stack()

            elif token == lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N:
                # \n escape `\`, `n`
                if self.__stream_stopped_with_leading_escape_character():
                    # push padding escape character `\` into JSON content
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                    # write current token symbol to JSON content
                    self.json_content.append(token_symbol)
                    # pop `\` from  stack
                    self.__pop_token_stack()
                    continue

                # check if json stream stopped with padding content, like case `[true , n`
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()

                # write current token symbol to JSON content
                self.json_content.append(token_symbol)

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

                # push `n`
                self.__push_token_stack(token)
                if self.__stream_stopped_in_an_array():
                    # in array, push `u`, `l`, `l`
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U
                    )
                else:
                    # in object, pop `n`
                    self.__pop_mirror_token_stack()

            elif token == lexer_tokens.TOKEN_ALPHABET_LOWERCASE_R:
                # \r escape `\`, `r`
                if self.__stream_stopped_with_leading_escape_character():
                    # push padding escape character `\` into JSON content
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                    # write current token symbol to JSON content
                    self.json_content.append(token_symbol)
                    # pop `\` from  stack
                    self.__pop_token_stack()
                    continue

                # write current token symbol to JSON content
                self.json_content.append(token_symbol)

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

                # check if `t` in token stack and `r`, `u`, `e in mirror stack
                def it_is_part_of_token_true():
                    left = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_T,
                    ]
                    right = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_R,
                    ]
                    return lexer_helper.match_stack(
                        self.token_stack, left
                    ) and lexer_helper.match_stack(self.mirror_token_stack, right)

                if not it_is_part_of_token_true():
                    continue

                self.__push_token_stack(token)
                self.__pop_mirror_token_stack()

            elif token == lexer_tokens.TOKEN_ALPHABET_LOWERCASE_S:
                # write current token symbol to JSON content
                self.json_content.append(token_symbol)

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

                # check if `f`, `a`, `l` in token stack and `s`, `e in mirror stack
                def it_is_part_of_token_false():
                    left = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_F,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_A,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
                    ]
                    right = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_S,
                    ]
                    return lexer_helper.match_stack(
                        self.token_stack, left
                    ) and lexer_helper.match_stack(self.mirror_token_stack, right)

                if not it_is_part_of_token_false():
                    continue

                self.__push_token_stack(token)
                self.__pop_mirror_token_stack()

            elif token == lexer_tokens.TOKEN_ALPHABET_LOWERCASE_T:
                # \t escape `\`, `t`
                if self.__stream_stopped_with_leading_escape_character():
                    # push padding escape character `\` into JSON content
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                    # write current token symbol to JSON content
                    self.json_content.append(token_symbol)
                    # pop `\` from  stack
                    self.__pop_token_stack()
                    continue

                # check if json stream stopped with padding content, like case `[true , t`
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()

                # write current token symbol to JSON content
                self.json_content.append(token_symbol)

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

                # push `t` to stack
                self.__push_token_stack(token)
                if self.__stream_stopped_in_an_array():
                    # in array
                    # push `r`, `u`, `e`
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_R
                    )
                else:
                    # in object
                    # pop `n`, `u`, `l`, `l`
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    # push `r`, `u`, `e`
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U
                    )
                    self.__push_mirror_token_stack(
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_R
                    )

            elif token == lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U:
                # unicode escape `\`, `u`
                if self.__stream_stopped_with_leading_escape_character():
                    self.__push_token_stack(token)
                    self.padding_content.append(token_symbol)
                    continue

                # write current token symbol to JSON content
                self.json_content.append(token_symbol)

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

                # check if `t`, `r` in token stack and, `u`, `e` in mirror stack
                def it_is_part_of_token_true():
                    left = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_T,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_R,
                    ]
                    right = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_E,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
                    ]
                    return lexer_helper.match_stack(
                        self.token_stack, left
                    ) and lexer_helper.match_stack(self.mirror_token_stack, right)

                # check if `n` in token stack and `u`, `l`, `l` in mirror stack
                def it_is_part_of_token_null():
                    left = [lexer_tokens.TOKEN_ALPHABET_LOWERCASE_N]
                    right = [
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_L,
                        lexer_tokens.TOKEN_ALPHABET_LOWERCASE_U,
                    ]
                    return lexer_helper.match_stack(
                        self.token_stack, left
                    ) and lexer_helper.match_stack(self.mirror_token_stack, right)

                if not it_is_part_of_token_true() and not it_is_part_of_token_null():
                    continue

                self.__push_token_stack(token)
                self.__pop_mirror_token_stack()

            elif token in [
                lexer_tokens.TOKEN_ALPHABET_UPPERCASE_A,
                lexer_tokens.TOKEN_ALPHABET_UPPERCASE_B,
                lexer_tokens.TOKEN_ALPHABET_UPPERCASE_C,
                lexer_tokens.TOKEN_ALPHABET_UPPERCASE_D,
                lexer_tokens.TOKEN_ALPHABET_LOWERCASE_C,
                lexer_tokens.TOKEN_ALPHABET_LOWERCASE_D,
                lexer_tokens.TOKEN_ALPHABET_UPPERCASE_F,
            ]:
                # as hex in unicode
                if self.__stream_stopped_in_an_string_unicode_escape():
                    self.__push_byte_into_padding_content(token_symbol)
                    # check if unicode escape is full length
                    if len(self.padding_content) == 6:
                        self.__append_padding_content_to_json_content()
                        self.__clean_padding_content()
                        # pop `\`, `u` from stack
                        self.__pop_token_stack()
                        self.__pop_token_stack()
                    continue

                # write current token symbol to JSON content
                self.json_content.append(token_symbol)

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

            elif token == lexer_tokens.TOKEN_ALPHABET_UPPERCASE_E:
                # as hex in unicode
                if self.__stream_stopped_in_an_string_unicode_escape():
                    self.__push_byte_into_padding_content(token_symbol)
                    # check if unicode escape is full length
                    if len(self.padding_content) == 6:
                        self.__append_padding_content_to_json_content()
                        self.__clean_padding_content()
                        # pop `\`, `u` from stack
                        self.__pop_token_stack()
                        self.__pop_token_stack()
                    continue

                # check if in a number, as `E` (exponent) in scientific notation
                if self.__stream_stopped_in_a_number_decimal_part_middle():
                    self.__push_byte_into_padding_content(token_symbol)
                    continue

                # write current token symbol to JSON content
                self.json_content.append(token_symbol)

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue
            elif token in [
                lexer_tokens.TOKEN_NUMBER_0,
                lexer_tokens.TOKEN_NUMBER_1,
                lexer_tokens.TOKEN_NUMBER_2,
                lexer_tokens.TOKEN_NUMBER_3,
                lexer_tokens.TOKEN_NUMBER_4,
                lexer_tokens.TOKEN_NUMBER_5,
                lexer_tokens.TOKEN_NUMBER_6,
                lexer_tokens.TOKEN_NUMBER_7,
                lexer_tokens.TOKEN_NUMBER_8,
                lexer_tokens.TOKEN_NUMBER_9,
            ]:
                # as number in unicode
                if self.__stream_stopped_in_an_string_unicode_escape():
                    self.__push_byte_into_padding_content(token_symbol)
                    # check if unicode escape is full length
                    if len(self.padding_content) == 6:
                        self.__append_padding_content_to_json_content()
                        self.__clean_padding_content()
                        #  pop `\`, `u` from stack
                        self.__pop_token_stack()
                        self.__pop_token_stack()
                    continue

                # check if json stream stopped with padding content, like `[1 , 1`
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()

                # in negative part of a number
                if self.__stream_stopped_in_a_negative_number_value_start():
                    self.__push_negative_into_json_content()
                    # pop `0` from mirror stack
                    self.__pop_mirror_token_stack()

                # write current token symbol to JSON content
                self.json_content.append(token_symbol)

                # in a string or a number, just skip token
                if (
                    self.__stream_stopped_in_a_string()
                    or self.__stream_stopped_in_a_number()
                ):
                    continue

                # in decimal part of a number
                if self.__stream_stopped_in_a_number_decimal_part():
                    self.__push_token_stack(lexer_tokens.TOKEN_NUMBER)
                    # pop placeholder `0` in decimal part
                    self.__pop_mirror_token_stack()
                    continue

                # first number type token, push token into stack
                self.__push_token_stack(lexer_tokens.TOKEN_NUMBER)

                # check if we are in an object or an array
                if self.__stream_stopped_in_an_array():
                    continue
                elif self.__stream_stopped_in_an_object_null_value_placeholder_start():
                    # pop `n`, `u`, `l`, `l`
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()

            elif token == lexer_tokens.TOKEN_COMMA:
                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    self.json_content.append(token_symbol)
                    continue
                # in a object or a array, keep the comma in stack but not write it into JSONContent, until next token arrival
                # the comma must following with token: quote, null, true, false, number
                self.__push_byte_into_padding_content(token_symbol)
                self.__push_token_stack(token)
            elif token == lexer_tokens.TOKEN_DOT:
                # write current token symbol to JSON content
                self.json_content.append(token_symbol)

                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    continue

                # use 0 for decimal part place holder
                self.__push_token_stack(token)
                self.__push_mirror_token_stack(lexer_tokens.TOKEN_NUMBER_0)

            elif token == lexer_tokens.TOKEN_SLASH:
                #  escape character `\`, `/`
                if self.__stream_stopped_with_leading_escape_character():
                    # push padding escape character `\` into JSON content
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                    # write current token symbol to JSON content
                    self.json_content.append(token_symbol)
                    # pop `\` from  stack
                    self.__pop_token_stack()
                    continue
                elif self.__stream_stopped_in_a_string():
                    self.json_content.append(token_symbol)
                    continue

            elif token == lexer_tokens.TOKEN_ESCAPE_CHARACTER:
                # double escape character `\`, `\`
                if self.__stream_stopped_with_leading_escape_character():
                    # push padding escape character `\` into JSON content
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()
                    # write current token symbol to JSON content
                    self.json_content.append(token_symbol)
                    # pop `\` from  stack
                    self.__pop_token_stack()
                    continue

                # just write escape character into stack and waitting other token trigger escape method.
                self.__push_token_stack(token)
                self.__push_byte_into_padding_content(
                    lexer_tokens.TOKEN_ESCAPE_CHARACTER_SYMBOL
                )
            elif token == lexer_tokens.TOKEN_NEGATIVE:
                # in a string, just skip token
                if self.__stream_stopped_in_a_string():
                    self.json_content.append(token_symbol)
                    continue

                # check if json stream stopped with padding content, like `[1 , -`
                if self.__have_padding_content():
                    self.__append_padding_content_to_json_content()
                    self.__clean_padding_content()

                # just write negative character into stack and waitting other token trigger it.
                self.__push_token_stack(token)
                if self.__stream_stopped_in_an_object_negative_number_value_start():
                    # pop `n`, `u`, `l`, `l` from mirror stack
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()
                    self.__pop_mirror_token_stack()

                # push `0` into mirror stack for placeholder
                self.__push_mirror_token_stack(lexer_tokens.TOKEN_NUMBER_0)

            else:
                return f"Unexpected token: {token}, token symbol: {token_symbol}"

            if token == lexer_tokens.TOKEN_EOF:
                break

        return None

    def complete_json(self) -> str:
        """
        complete the incomplete JSON string by concat JSON content and mirror tokens
        """
        # This combines json_content and mirror token stack into a complete JSON string
        return "".join(self.json_content) + self.__dump_mirror_token_stack_to_string()



#import nest_asyncio
try:
    import readline
except ImportError:
    import pyreadline as readline
from .Request import Request
from .Agent import AgentFactory
from .Facility import FacilityManager
from .WebSocket import WebSocketServer, WebSocketClient
from .Workflow import Workflow, Schema as WorkflowSchema, Checkpoint as WorkflowCheckpoint
from .AppConnector import AppConnector
from ._global import global_plugin_manager, global_settings, global_storage, global_tool_manager, global_websocket_server
from .utils import *

#nest_asyncio.apply()
def create_agent(*args, **kwargs):
    return AgentFactory().create_agent(*args, **kwargs)

facility = FacilityManager()
lib = facility

set_settings = global_settings.set

def set_compatible(compatible_version: str):
    if isinstance(compatible_version, str):
        compatible_version = int(compatible_version.replace(".", ""))
    set_settings("compatible_version", compatible_version)
    # auto switch settings by compatible version
    if compatible_version <= 3320:
        set_settings("storage_type", "FileStorage")
        set_settings("storage.FileStorage.path", "./.Agently")

def register_plugin(module_name:str, plugin_name: str, plugin: callable):
    global_plugin_manager.register(module_name, plugin_name, plugin)
    facility.refresh_plugins()

def set_plugin_settings(module_name: str, plugin_name: str, key: str, value: any):
    global_plugin_manager.set_settings(f"plugin_settings.{ module_name }.{ plugin_name }", key, value)

def attach_workflow(name: str, workflow: object):
    class AttachedWorkflow:
        def __init__(self, agent: object):
            self.agent = agent
            self.get_debug_status = lambda: self.agent.settings.get_trace_back("is_debug")
            self.settings = RuntimeCtxNamespace(f"plugin_settings.agent_component.{ name }", self.agent.settings)
        
        def start_workflow(self, init_inputs: dict=None, init_storage: dict={}):
            if not isinstance(init_storage, dict):
                raise Exception("[Workflow] Initial storage must be a dict.")
            init_storage.update({ "$agent": self.agent })
            return workflow.start(init_inputs, storage=init_storage)
        
        def export(self):
            return {
                "alias": {
                    name: { "func": self.start_workflow, "return_value": True },
                }
            }
    return register_plugin("agent_component", name, AttachedWorkflow)


import time as time_raw
from typing import TYPE_CHECKING, Union
from ..utils.chunk_helper import deep_copy_simply

if TYPE_CHECKING:
    from .State import RuntimeState

class Snapshot:
  """主要负责单个 RuntimeState 的静态化挂载及管理逻辑"""

  def __init__(self, name: str, state: Union['RuntimeState', dict], time: int = None) -> None:
    self.name = name
    self.state_schema = state if isinstance(state, dict) else state.export()
    self.time = time or time_raw.time()
  
  def get_state_val(self, keys_with_dots: str):
    """获取指定状态值"""
    keys = keys_with_dots.split('.')
    pointer = self.state_schema
    for key in keys:
      if not pointer or key not in pointer:
        return None
      pointer = pointer.get(key)
    return deep_copy_simply(pointer)

  def export(self) -> dict:
     return deep_copy_simply({
        'name': self.name,
        'state': self.state_schema,
        'time': self.time
     })


#ABC = Abstract Base Class
from abc import ABC, abstractmethod

class ToolABC(ABC):
    def __init__(self, tool_manager: object):
        self.tool_manager = tool_manager

    def export(self):
        return {
            "<tool_name>": {
                "desc": "<tool_desc>",
                "args": {
                    "<arg_1>": ("<arg_type>", "<arg_desc>"),
                },
                "func": callable,
            },
        }


from .utils import ComponentABC

class ReplyReformer(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.reform_handler = None

    def add_reply_reform_handler(self, reform_handler: callable):
        self.reform_handler = reform_handler
        return self.agent

    def _suffix(self, event, data):
        if event == "response:finally" and callable(self.reform_handler):
            self.agent.request.response_cache["reply"] = self.reform_handler(data)

    def export(self):
        return {
            "prefix": None,
            "suffix": self._suffix,
            "alias": {
                "reform_reply": { "func": self.add_reply_reform_handler }
            },
        }

def export():
    return ("ReplyReformer", ReplyReformer)